import torch

from ...manager import BaseManager


def attach_soteria_to_client(
    cls,
    input_layer,
    perturbed_layer,
    epsilon=0.2,
    target_layer_name=None,
):
    class SoteriaClientWrapper(cls):
        """Implementation of https://arxiv.org/pdf/2012.06043.pdf"""

        def __init__(self, *args, **kwargs):
            super(SoteriaClientWrapper, self).__init__(*args, **kwargs)
            self.input_layer = input_layer
            self.perturbed_layer = perturbed_layer
            self.epsilon = epsilon
            self.target_layer_name = target_layer_name

            self._set_hook()

        def _set_hook(self):
            self.inputs = {}
            self.outputs = {}

            def get_input(name):
                def hook(model, x, output):
                    if not x[0].requires_grad:
                        raise ValueError("x.requires_grad should be True")
                    self.inputs[name] = x[0]

                return hook

            def get_output(name):
                def hook(model, x, output):
                    self.outputs[name] = output

                return hook

            getattr(self.model, self.input_layer).register_forward_hook(
                get_input(self.input_layer)
            )
            getattr(self.model, self.perturbed_layer).register_forward_hook(
                get_output(self.perturbed_layer)
            )

        def action_before_lossbackward(self):
            input_data = self.inputs[self.input_layer]
            feature = self.outputs[self.perturbed_layer]

            mask = torch.zeros_like(feature)
            r_dfr_dx_norm = torch.zeros_like(feature)

            rep_size = feature.shape[1]
            for i in range(rep_size):
                mask[:, i] = 1
                feature.backward(
                    mask, retain_graph=True
                )  # culc the derivative of feature_2 @ df_dtarget
                dfri_dx = input_data.grad.data
                r_dfr_dx_norm[:, i] = feature[:, i] / torch.norm(
                    dfri_dx.view(dfri_dx.shape[0], -1), dim=1
                )

                self.model.zero_grad()
                input_data.grad.data.zero_()
                mask[:, i] = 0

            self.topk_idxs = torch.topk(
                r_dfr_dx_norm.mean(dim=0), int(rep_size * self.epsilon)
            )[1]

        def action_after_lossbackward(self, target_layer_name=None):
            target_layer_name = (
                f"{self.perturbed_layer}.weight"
                if target_layer_name is None
                else target_layer_name
            )
            dl_dw = {
                layer_name: params.grad
                for layer_name, params in zip(
                    self.model.state_dict(), self.model.parameters()
                )
            }
            dl_dw[target_layer_name][self.topk_idxs, :] = 0

        def backward(self, loss):
            self.action_before_lossbackward()
            super().backward(loss)
            self.action_after_lossbackward(self.target_layer_name)

        def local_train(
            self, local_epoch, criterion, trainloader, optimizer, communication_id=0
        ):
            for i in range(local_epoch):
                running_loss = 0.0
                running_data_num = 0
                for _, data in enumerate(trainloader, 0):
                    inputs, labels = data
                    inputs = inputs.to(self.device)
                    inputs.retain_grad()
                    labels = labels.to(self.device)

                    optimizer.zero_grad()
                    self.zero_grad()

                    outputs = self(inputs)
                    loss = criterion(outputs, labels)
                    self.backward(loss)

                    optimizer.step()

                    running_loss += loss.item()
                    running_data_num += inputs.shape[0]

                print(
                    f"communication {communication_id}, epoch {i}: client-{self.user_id+1}",
                    running_loss / running_data_num,
                )

    return SoteriaClientWrapper


class SoteriaClientManager(BaseManager):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        return attach_soteria_to_client(cls, *self.args, **self.kwargs)



from numpy.polynomial import Polynomial

from .plaintext import CKKSPlaintext
from .utils import polydiv_coef


class CKKSCiphertext:
    def __init__(self, c0: Polynomial, c1: Polynomial, N: int, q: int, scale: int):
        self.c0 = c0
        self.c1 = c1
        self.N = N
        self.q = q
        self.scale = scale

        self.poly_modulo = Polynomial(
            [1 if i == 0 or i == N else 0 for i in range(N + 1)]
        )

    def __add__(self, other):
        if type(other) == CKKSPlaintext:
            return CKKSCiphertext(
                polydiv_coef(self.c0 + other.p, self.q) % self.poly_modulo,
                polydiv_coef(self.c1 + other.p, self.q) % self.poly_modulo,
                self.N,
                self.q,
                self.scale,
            )
        elif type(other) == CKKSCiphertext:
            return CKKSCiphertext(
                polydiv_coef(self.c0 + other.c0, self.q) % self.poly_modulo,
                polydiv_coef(self.c1 + other.c1, self.q) % self.poly_modulo,
                self.N,
                self.q,
                self.scale,
            )
        else:
            raise TypeError("error!")

    def __mul__(self, other):
        if type(other) == CKKSPlaintext:
            return CKKSCiphertext(
                polydiv_coef((self.c0 * other.p) / self.scale, self.q)
                % self.poly_modulo,
                polydiv_coef((self.c1 * other.p) / self.scale, self.q)
                % self.poly_modulo,
                self.N,
                self.q,
                self.scale,
            )
        else:
            raise TypeError("error!")



from typing import List

from ..base_attack import BaseAttacker


class Generator_Attack(BaseAttacker):
    def __init__(
        self,
        target_model,
        attacker_model,
        attacker_optimizer,
        log_interval=1,
        early_stopping=5,
        device="cpu",
    ):
        if type(target_model) == List:
            super().__init__(target_model=target_model[0])
        else:
            super().__init__(target_model=target_model)
        self.attacker_model = attacker_model
        self.attacker_optimizer = attacker_optimizer
        self.log_interval = log_interval
        self.early_stopping = early_stopping
        self.device = device

        self.target_model_list = (
            target_model if type(target_model) == list else [target_model]
        )

    def culc_loss(self, dataloader, x_pos=0, y_pos=1, arbitrary_y=False):
        running_loss = 0
        for data in dataloader:
            x = data[x_pos]
            x = x.to(self.device)

            loss = 0
            for target_model in self.target_model_list:
                target_outputs = data[y_pos] if arbitrary_y else target_model(x)
                target_outputs = target_outputs.to(self.device)

                attack_outputs = self.attacker_model(target_outputs)
                loss = loss + ((x - attack_outputs) ** 2).mean()

            running_loss = running_loss + loss / len(dataloader)

        return running_loss

    def fit(self, dataloader, epoch, x_pos=0, y_pos=1, arbitrary_y=False):
        best_loss = float("inf")
        best_epoch = 0
        for i in range(epoch):
            running_loss = 0
            for data in dataloader:
                x = data[x_pos]
                x = x.to(self.device)

                loss = 0
                self.attacker_optimizer.zero_grad()
                for target_model in self.target_model_list:
                    target_outputs = data[y_pos] if arbitrary_y else target_model(x)
                    target_outputs = target_outputs.to(self.device)

                    attack_outputs = self.attacker_model(target_outputs)
                    loss = loss + ((x - attack_outputs) ** 2).mean() / len(
                        self.target_model_list
                    )
                loss.backward()
                self.attacker_optimizer.step()

                running_loss = running_loss + loss.item() / len(dataloader)

            if self.log_interval != 0 and i % self.log_interval == 0:
                print(f"epoch {i}: reconstruction_loss {running_loss}")

            if running_loss < best_loss:
                best_loss = running_loss
                best_epoch = i
            else:
                if i - best_epoch > self.early_stopping:
                    break

    def attack(self, data_tensor):
        return self.attacker_model(data_tensor)



import torch
from torch.utils.data import TensorDataset

from ....manager import BaseManager


def attach_dpsgd_to_client(cls, privacy_manager, sigma):
    dpoptimizer_wrapper, lot_loader, batch_loader = privacy_manager.privatize(sigma)

    class DPSGDClientWrapper(cls):
        def __init__(self, *args, **kwargs):
            super(DPSGDClientWrapper, self).__init__(*args, **kwargs)
            self.privacy_manager = privacy_manager

        def local_train(
            self, local_epoch, criterion, trainloader, optimizer, communication_id=0
        ):
            _ = trainloader  # we do not explicitly use it

            for i in range(local_epoch):
                running_loss = 0.0
                running_data_num = 0
                for X_lot, y_lot in lot_loader(optimizer):
                    for X_batch, y_batch in batch_loader(TensorDataset(X_lot, y_lot)):
                        optimizer.zero_grad()
                        pred = self(X_batch)
                        loss = criterion(pred, y_batch.to(torch.int64))
                        loss.backward()
                        optimizer.step()

                        running_loss += loss.item()
                        running_data_num += X_batch.shape[0]

                print(
                    f"communication {communication_id}, epoch {i}: client-{self.user_id+1}",
                    running_loss / running_data_num,
                )

    return DPSGDClientWrapper, dpoptimizer_wrapper


class DPSGDClientManager(BaseManager):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        return attach_dpsgd_to_client(cls, *self.args, **self.kwargs)



from ..core import BaseClient


class SplitNNClient(BaseClient):
    def __init__(self, model, user_id=0):
        super().__init__(model, user_id=user_id)
        self.own_intermidiate = None
        self.prev_intermidiate = None
        self.grad_from_next_client = None

    def forward(self, prev_intermediate):
        """Send intermidiate tensor to the server

        Args:
            x (torch.Tensor): the input data

        Returns:
            intermidiate_to_next_client (torch.Tensor): the output of client-side
                                                   model which the client sent
                                                   to the server
        """
        self.prev_intermidiate = prev_intermediate
        self.own_intermidiate = self.model(prev_intermediate)
        intermidiate_to_next_client = self.own_intermidiate.detach().requires_grad_()
        return intermidiate_to_next_client

    def upload(self, x):
        return self.forward(x)

    def download(self, grad_from_next_client):
        self._client_backward(grad_from_next_client)

    def _client_backward(self, grad_from_next_client):
        """Client-side back propagation

        Args:
            grad_from_server: gradient which the server send to the client
        """
        self.grad_from_next_client = grad_from_next_client
        self.own_intermidiate.backward(grad_from_next_client)

    def distribute(self):
        return self.prev_intermidiate.grad.clone()



import numpy as np
from numpy.polynomial import Polynomial

from .plaintext import CKKSPlaintext
from .utils import coordinate_wise_random_rounding


class CKKSEncoder:
    """Basic CKKS encoder to encode complex vectors into polynomials.
    This class is mainly based on https://blog.openmined.org/ckks-explained-part-1-simple-encoding-and-decoding/
    """

    def __init__(self, M: int, scale: float):
        """Initialization of the encoder for M a power of 2.

        Args:
            M:
            scale:

        Attributed:
            xi: M-th root of unity used as a basis
            N:
            A:
            sigma_R_basis
        """
        self.M = M
        self.scale = scale

        self.xi = np.exp(2 * np.pi * 1j / M)
        self.N = M // 2
        self.A = self._vandermonde()
        self.sigma_R_basis = self.A.T

    def _vandermonde(self) -> np.array:
        """Get vandermonde matrix of a m-th root of unity.

        Returns:
            np.array: a vandermonde matrix (N^N)
        """
        roots = np.array([self.xi ** (2 * i + 1) for i in range(self.N)])
        matrix = np.array([roots**j for j in range(self.N)])
        return matrix.T

    def sigma_inverse(self, z: np.array) -> Polynomial:
        """Encodes the vector z (C^{N/2}) to a polinomial ().

        Args:
            z (np.array): target vector

        Returns:
            Polynomial: the encoded polinoimial
        """
        return Polynomial(np.linalg.solve(self.A, z))

    def sigma(self, p: Polynomial) -> np.array:
        """Decodes a polynomial to a vector (C^{N/2})

        Args:
            p: encoded polinoimial

        Returns:
            np.array: decoded vector
        """
        return np.array([p(self.xi ** (2 * i + 1)) for i in range(self.N)])

    def pi(self, z: np.array) -> np.array:
        """Projects a vector of H into C^{N/2}.

        Args:
            z:

        Returns:
            np.array:
        """
        return z[: self.N // 2]

    def pi_inverse(self, z: np.array) -> np.array:
        """Expands a vector of C^{N/2} by expanding it with its
        complex conjugate."""

        z_conjugate = z[::-1]
        z_conjugate = [np.conjugate(x) for x in z_conjugate]
        return np.concatenate([z, z_conjugate])

    def compute_basis_coordinates(self, z):
        """Computes the coordinates of a vector with respect to the orthogonal lattice basis."""
        output = np.array(
            [np.real(np.vdot(z, b) / np.vdot(b, b)) for b in self.sigma_R_basis]
        )
        return output

    def sigma_R_discretization(self, z):
        """Projects a vector on the lattice using coordinate wise random rounding."""
        coordinates = self.compute_basis_coordinates(z)
        rounded_coordinates = coordinate_wise_random_rounding(coordinates)
        y = np.matmul(self.sigma_R_basis.T, rounded_coordinates)
        return y

    def encode(self, z: np.array) -> CKKSPlaintext:
        """Encodes a vector by expanding it first to H,
        scale it, project it on the lattice of sigma(R), and performs
        sigma inverse.
        """
        pi_z = self.pi_inverse(z)
        scaled_pi_z = self.scale * pi_z
        rounded_scale_pi_zi = self.sigma_R_discretization(scaled_pi_z)
        p = self.sigma_inverse(rounded_scale_pi_zi)

        # We round it afterwards due to numerical imprecision
        coef = np.round(np.real(p.coef)).astype(int)
        pt = CKKSPlaintext(Polynomial(coef), self.N, self.scale)
        return pt

    def decode(self, pt: CKKSPlaintext) -> np.array:
        """Decodes a polynomial by removing the scale,
        evaluating on the roots, and project it on C^(N/2)"""
        rescaled_p = pt.p / self.scale
        z = self.sigma(rescaled_p)
        pi_z = self.pi(z)
        return pi_z



import numpy as np
import torch
from torch.utils.data import DataLoader


class PoissonSampler:
    def __init__(self, dataset, lot_size, iterations):
        self.dataset_size = len(dataset)
        self.lot_size = lot_size
        self.iterations = iterations

    def __iter__(self):
        for _ in range(self.iterations):
            indices = np.where(
                torch.rand(self.dataset_size) < (self.lot_size / self.dataset_size)
            )[0]
            if indices.size > 0:
                yield indices

    def __len__(self):
        return self.iterations


class DPWrapperLotDataIterator:
    def __init__(self, original_iterator, dp_optimizer):
        self.original_iterator = original_iterator
        self.dp_optimizer = dp_optimizer
        self.init_flag = True

    def __iter__(self):
        return self

    def _reset(self, *args, **kwargs):
        return self.original_iterator._reset(*args, **kwargs)

    def _next_index(self, *args, **kwargs):
        return self.original_iterator._next_index(*args, **kwargs)

    def __next__(self):
        if not self.init_flag:
            self.dp_optimizer.step_for_lot()
        else:
            self.init_flag = False

        data = self.original_iterator.__next__()
        self.dp_optimizer.zero_grad_for_lot()

        return data

    def __len__(self):
        return self.original_iterator.__len__()

    def __getstate__(self):
        raise self.original_iterator.__getstate__()


class LotDataLoader(DataLoader):
    def __init__(self, dp_optimizer, *args, **kwargs):
        super(LotDataLoader, self).__init__(*args, **kwargs)
        self.dp_optimizer = dp_optimizer
        self.init_flag = True

    def __iter__(self):
        return DPWrapperLotDataIterator(
            super(LotDataLoader, self).__iter__(), self.dp_optimizer
        )



import torch


def total_variance(x):
    dx = torch.mean(torch.abs(x[:, :, :, :-1] - x[:, :, :, 1:]))
    dy = torch.mean(torch.abs(x[:, :, :-1, :] - x[:, :, 1:, :]))
    return dx + dy


def label_matching(pred, label):
    onehot_label = torch.eye(pred.shape[-1])[label]
    onehot_label = onehot_label.to(pred.device)
    return torch.sqrt(torch.sum((pred - onehot_label) ** 2))


def group_consistency(x, group_x):
    mean_group_x = sum(group_x) / len(group_x)
    return torch.norm(x - mean_group_x, p=2)


def bn_regularizer(feature_maps, bn_layers):
    bn_reg = 0
    for i, layer in enumerate(bn_layers):
        fm = feature_maps[i]
        if len(fm.shape) == 3:
            dim = [0, 2]
        elif len(fm.shape) == 4:
            dim = [0, 2, 3]
        elif len(fm.shape) == 5:
            dim = [0, 2, 3, 4]
        bn_reg += torch.norm(fm.mean(dim=dim) - layer.state_dict()["running_mean"], p=2)
        bn_reg += torch.norm(fm.var(dim=dim) - layer.state_dict()["running_var"], p=2)
    return bn_reg



import numpy as np
import torch
import torch.nn.functional as F

from ...manager import BaseManager

EPS = 1e-8


def attach_foolsgold_to_server(cls):
    """Wraps the given class in FoolsGoldServerWrapper.

    Returns:
        cls: a class wrapped in FoolsGoldServerWrapper
    """

    class FoolsGoldServerWrapper(cls):
        """Implementation of https://arxiv.org/abs/1808.04866"""

        def __init__(self, *args, **kwargs):
            super(FoolsGoldServerWrapper, self).__init__(*args, **kwargs)

            tmp_flatten_local_gradient = torch.cat(
                [p.view(-1) for p in self.server_model.parameters()]
            ).to(self.device)
            self.aggregate_historical_gradients = [
                torch.zeros_like(tmp_flatten_local_gradient)
                for i in range(len(self.clients))
            ]
            self.cs = np.zeros((len(self.clients), len(self.clients)))
            self.v = np.zeros(len(self.clients))
            self.alpha = np.zeros(len(self.clients))

        def update(self):
            self.update_weight()
            self.update_from_gradients(self.alpha)

        def update_weight(self):
            """Updates weight for each client given the received local gradients."""
            for i, local_gradient in enumerate(self.uploaded_gradients):
                self.aggregate_historical_gradients[i] += torch.cat(
                    [g.to(self.device).view(-1) for g in local_gradient[1]]
                ).to(self.device)

            num_clients = len(self.uploaded_gradients)

            for i_idx in range(num_clients):
                for j_idx in range(i + 1, num_clients):
                    self.cs[i_idx][j_idx] = F.cosine_similarity(
                        self.aggregate_historical_gradients[i_idx],
                        self.aggregate_historical_gradients[j_idx],
                        0,
                        EPS,
                    )
            self.v = np.max(self.cs, axis=1)

            for i_idx in range(num_clients):
                for j_idx in range(num_clients):
                    if i_idx == j_idx:
                        continue
                    if self.v[j_idx] > self.v[i_idx]:
                        self.cs[i_idx][j_idx] *= self.v[i_idx] / self.v[j_idx]

            self.alpha = np.max(self.cs, axis=1)
            self.alpha = self.alpha / (np.max(self.alpha) + EPS)

    return FoolsGoldServerWrapper


class FoolsGoldServerManager(BaseManager):
    """Manager class for FoolsGold proposed in https://arxiv.org/abs/1808.04866."""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        """Wraps the given class in FoolsGoldServerWrapper.

        Returns:
            cls: a class wrapped in FoolsGoldServerWrapper
        """
        return attach_foolsgold_to_server(cls, *self.args, **self.kwargs)



import torch

from .base import BaseFLOptimizer


class AdamFLOptimizer(BaseFLOptimizer):
    """Implementation of Adam to update the global model of Federated Learning.

    Args:
        parameters (List[torch.nn.Parameter]): parameters of the model
        lr (float, optional): learning rate. Defaults to 0.01.
        weight_decay (float, optional): coefficient of weight decay. Defaults to 0.0001.
        beta1 (float, optional): 1st-order exponential decay. Defaults to 0.9.
        beta2 (float, optional): 2nd-order exponential decay. Defaults to 0.999.
        epsilon (float, optional): a small value to prevent zero-devision. Defaults to 1e-8.
    """

    def __init__(
        self,
        parameters,
        lr=0.01,
        weight_decay=0.0001,
        beta1=0.9,
        beta2=0.999,
        epsilon=1e-8,
    ):
        super().__init__(parameters, lr=lr, weight_decay=weight_decay)
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon

        self.m = [torch.zeros_like(param.data) for param in self.parameters]
        self.v = [torch.zeros_like(param.data) for param in self.parameters]

    def step(self, grads):
        """Update the parameters with the give gradient

        Args:
            grads (List[torch.Tensor]): list of gradients
        """
        for i, (param, grad) in enumerate(zip(self.parameters, grads)):
            self.m[i] = self.beta1 * self.m[i] + (1 - self.beta1) * grad
            self.v[i] = self.beta2 * self.v[i] + (1 - self.beta2) * (grad * grad)
            m_hat = self.m[i] / (1 - self.beta1**self.t)
            v_hat = self.v[i] / (1 - self.beta2**self.t)
            param.data -= self.lr * (
                m_hat / torch.sqrt(v_hat) + self.weight_decay * param.data
            )
        self.t += 1



from abc import abstractmethod

import torch


class BaseClient(torch.nn.Module):
    """Abstract class foe the client of collaborative learning.

    Args:
        model (torch.nn.Module): a local model
        user_id (int, optional): id of this client. Defaults to 0.
    """

    def __init__(self, model, user_id=0):
        """Initialize BaseClient"""
        super(BaseClient, self).__init__()
        self.model = model
        self.user_id = user_id

    def forward(self, x):
        return self.model(x)

    def train(self):
        self.model.train()

    def eval(self):
        self.model.eval()

    def backward(self, loss):
        """Execute backward mode automatic differentiation with the give loss.

        Args:
            loss (torch.Tensor): the value of calculated loss.
        """
        loss.backward()

    @abstractmethod
    def upload(self):
        """Upload the locally learned informatino to the server."""
        pass

    @abstractmethod
    def download(self):
        """Download the global model from the server."""
        pass

    @abstractmethod
    def local_train(self):
        pass



import copy

from ..core.api import BaseFedAPI


class FedAVGAPI(BaseFedAPI):
    """Implementation of FedAVG (McMahan, Brendan, et al. 'Communication-efficient learning of deep networks from decentralized data.' Artificial intelligence and statistics. PMLR, 2017.)

    Args:
        server (FedAvgServer): FedAVG server.
        clients ([FedAvgClient]): a list of FedAVG clients.
        criterion (function): loss function.
        local_optimizers ([torch.optimizer]): a list of local optimizers for clients
        local_dataloaders ([toch.dataloader]): a list of local dataloaders for clients
        num_communication (int, optional): number of communication. Defaults to 1.
        local_epoch (int, optional): number of epochs for local training within each communication. Defaults to 1.
        use_gradients (bool, optional): communicate gradients if True. Otherwise communicate parameters. Defaults to True.
        custom_action (function, optional): arbitrary function that takes this instance itself. Defaults to lambdax:x.
        device (str, optional): device type. Defaults to "cpu".
    """

    def __init__(
        self,
        server,
        clients,
        criterion,
        local_optimizers,
        local_dataloaders,
        num_communication=1,
        local_epoch=1,
        use_gradients=True,
        custom_action=lambda x: x,
        device="cpu",
    ):
        self.server = server
        self.clients = clients
        self.criterion = criterion
        self.local_optimizers = local_optimizers
        self.local_dataloaders = local_dataloaders
        self.num_communication = num_communication
        self.local_epoch = local_epoch
        self.use_gradients = use_gradients
        self.custom_action = custom_action
        self.device = device

        self.client_num = len(self.clients)

        local_dataset_sizes = [
            len(dataloader.dataset) for dataloader in self.local_dataloaders
        ]
        sum_local_dataset_sizes = sum(local_dataset_sizes)
        self.clients_weight = [
            dataset_size / sum_local_dataset_sizes
            for dataset_size in local_dataset_sizes
        ]

    def local_train(self, i):
        for client_idx in range(self.client_num):
            self.clients[client_idx].local_train(
                self.local_epoch,
                self.criterion,
                self.local_dataloaders[client_idx],
                self.local_optimizers[client_idx],
                communication_id=i,
            )

    def run(self):
        self.server.force_send_model_state_dict = True
        self.server.distribute()
        self.server.force_send_model_state_dict = False

        for i in range(self.num_communication):
            self.local_train(i)
            self.server.receive(use_gradients=self.use_gradients)
            if self.use_gradients:
                self.server.update_from_gradients(weight=self.clients_weight)
            else:
                self.server.update_from_parameters(weight=self.clients_weight)
            self.server.distribute()

            self.custom_action(self)


class MPIFedAVGAPI(BaseFedAPI):
    def __init__(
        self,
        comm,
        party,
        is_server,
        criterion,
        local_optimizer=None,
        local_dataloader=None,
        num_communication=1,
        local_epoch=1,
        custom_action=lambda x: x,
        device="cpu",
    ):
        self.comm = comm
        self.party = party
        self.is_server = is_server
        self.criterion = criterion
        self.local_optimizer = local_optimizer
        self.local_dataloader = local_dataloader
        self.num_communication = num_communication
        self.local_epoch = local_epoch
        self.custom_action = custom_action
        self.device = device

    def run(self):
        self.party.mpi_initialize()
        self.comm.Barrier()

        for i in range(self.num_communication):
            if not self.is_server:
                self.local_train(i)
            self.party.action()

            self.custom_action(self)
            self.comm.Barrier()

    def local_train(self, com_cnt):
        self.party.prev_parameters = []
        for param in self.party.model.parameters():
            self.party.prev_parameters.append(copy.deepcopy(param))

        self.party.local_train(
            self.local_epoch,
            self.criterion,
            self.local_dataloader,
            self.local_optimizer,
            communication_id=com_cnt,
        )



import functools

import numpy as np
import torch

HANDLED_FUNCTIONS = {}


def implements(torch_function):
    """Registers a torch function override for PaillierTensor"""

    @functools.wraps(torch_function)
    def decorator(func):
        HANDLED_FUNCTIONS[torch_function] = func
        return func

    return decorator


class PaillierTensor(object):
    """torch.Tensor-like object for Paillier Encryption"""

    def __init__(self, paillier_array):
        if type(paillier_array) == list:
            self._paillier_np_array = np.array(paillier_array)
        elif type(paillier_array) == np.ndarray:
            self._paillier_np_array = paillier_array
        else:
            raise TypeError(f"{type(paillier_array)} is not supported.")

    def __repr__(self):
        return "PaillierTensor"

    def decrypt(self, sk, device="cpu"):
        return torch.Tensor(
            np.vectorize(lambda x: sk.decrypt2float(x))(self._paillier_np_array)
        ).to(device)

    def tensor(self, sk=None):
        if sk is not None:
            return self.decypt(sk)
        else:
            return torch.zeros(self._paillier_np_array.shape)

    def numpy(self):
        return self._paillier_np_array

    def detach(self):
        return self

    def cpu(self):
        return self

    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        if func not in HANDLED_FUNCTIONS or not all(
            issubclass(t, (torch.Tensor, PaillierTensor)) for t in types
        ):
            return NotImplemented
        return HANDLED_FUNCTIONS[func](*args, **kwargs)

    @implements(torch.add)
    def add(input, other):
        if type(other) in [int, float]:
            return PaillierTensor(input._paillier_np_array + other)
        elif type(other) in [torch.Tensor, PaillierTensor]:
            return PaillierTensor(
                input._paillier_np_array + other.detach().cpu().numpy()
            )
        else:
            raise NotImplementedError(f"{type(other)} is not supported.")

    @implements(torch.sub)
    def sub(input, other):
        if type(other) in [int, float]:
            return PaillierTensor(input._paillier_np_array + (-1) * other)
        elif type(other) in [torch.Tensor, PaillierTensor]:
            return PaillierTensor(
                input._paillier_np_array + (-1) * other.detach().cpu().numpy()
            )
        else:
            raise NotImplementedError(f"{type(other)} is not supported.")

    @implements(torch.mul)
    def mul(input, other):
        if type(other) in [int, float]:
            return PaillierTensor(input._paillier_np_array * other)
        elif type(other) in [torch.Tensor, PaillierTensor]:
            return PaillierTensor(
                input._paillier_np_array * other.detach().cpu().numpy()
            )
        else:
            raise NotImplementedError(f"{type(other)} is not supported.")

    def __add__(self, other):
        return torch.add(self, other)

    def __iadd__(self, other):
        self = torch.add(self, other)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return torch.sub(self, other)

    def __isub__(self, other):
        self = torch.sub(self, other)

    def __rsub__(self, other):
        return self.__sub__(other)

    def __mul__(self, other):
        return torch.mul(self, other)

    def __rmul__(self, other):
        return self.__mul__(other)



from abc import abstractmethod

import torch


class BaseServer(torch.nn.Module):
    """Abstract class for the server of the collaborative learning.

    Args:
        clients (List[BaseClient]): a list of clients
        server_model (torch.nn.Module): a global model
        server_id (int, optional): the id of this server. Defaults to 0.
    """

    def __init__(self, clients, server_model, server_id=0):
        """Initialie BaseServer"""
        super(BaseServer, self).__init__()
        self.clients = clients
        self.server_id = server_id
        self.server_model = server_model
        self.num_clients = len(clients)

    def forward(self, x):
        return self.server_model(x)

    def train(self):
        self.server_model.train()

    def eval(self):
        self.server_model.eval()

    @abstractmethod
    def action(self):
        """Execute thr routine of each communication."""
        pass

    @abstractmethod
    def update(self):
        """Update the global model."""
        pass

    @abstractmethod
    def distribute(self):
        """Distribute the global model to each client."""
        pass



from ..core.api import BaseFedAPI


class SplitNNAPI(BaseFedAPI):
    def __init__(self, clients, optimizers, dataloader, criterion, num_epoch):
        super().__init__()
        self.clients = clients
        self.optimizers = optimizers
        self.dataloader = dataloader
        self.criterion = criterion
        self.num_epoch = num_epoch

        self.num_clients = len(clients)
        self.recent_output = None
        self.loss_log = []

    def local_train(self):
        for data in self.dataloader:
            self.zero_grad()
            inputs, labels = data
            outputs = self(inputs)
            loss = self.criterion(outputs, labels)
            self.backward(loss)
            self.step()

            self.loss_log.append(loss.item())

    def run(self):
        self.train()

        for _ in range(self.num_epoch):
            self.local_train()

    def __call__(self, *args, **kwds):
        return self.forward(*args, **kwds)

    def forward(self, x):
        intermidiate_to_next_client = x
        for client in self.clients:
            intermidiate_to_next_client = client.upload(intermidiate_to_next_client)
        output = intermidiate_to_next_client
        self.recent_output = output
        return output

    def backward(self, loss):
        loss.backward()
        return self.backward_gradient(self.recent_output.grad)

    def backward_gradient(self, grads_outputs):
        grad_from_next_client = grads_outputs
        for i in range(self.num_clients - 1, -1, -1):
            self.clients[i].download(grad_from_next_client)
            if i != 0:
                grad_from_next_client = self.clients[i].distribute()
        return grad_from_next_client

    def train(self):
        for client in self.clients:
            client.train()

    def eval(self):
        for client in self.clients:
            client.train()

    def zero_grad(self):
        for opt in self.optimizers:
            opt.zero_grad()

    def step(self):
        for opt in self.optimizers:
            opt.step()



from torch.utils.data import DataLoader

from .dataloader import LotDataLoader, PoissonSampler
from .optimizer import attach_differential_privacy_mechanism


class DPSGDManager:
    def __init__(
        self,
        accountant,
        optimizer_cls,
        l2_norm_clip,
        dataset,
        lot_size,
        batch_size,
        iterations,
    ):
        self.accountant = accountant
        self.optimizer_cls = optimizer_cls
        self.l2_norm_clip = l2_norm_clip
        self.dataset = dataset
        self.lot_size = lot_size
        self.batch_size = batch_size
        self.iterations = iterations

    def privatize(self, noise_multiplier):
        dpoptimizer_class = attach_differential_privacy_mechanism(
            self.optimizer_cls,
            self.accountant,
            self.l2_norm_clip,
            noise_multiplier,
            self.lot_size,
            self.batch_size,
            len(self.dataset),
        )

        def lot_loader(dp_optimizer):
            return LotDataLoader(
                dp_optimizer,
                self.dataset,
                batch_sampler=PoissonSampler(
                    self.dataset, self.lot_size, self.iterations
                ),
            )

        def batch_loader(lot):
            return DataLoader(lot, batch_size=self.batch_size, drop_last=True)

        return dpoptimizer_class, lot_loader, batch_loader



from abc import ABCMeta, abstractmethod


class BaseManager(metaclass=ABCMeta):
    """Abstract class for Manager API"""

    def __init__(self):
        pass

    @abstractmethod
    def attach(self):
        pass



import torch.nn.functional as F

from ..fedavg import FedAVGClient


class FedKDClient(FedAVGClient):
    """Implementation of FedKD (https://arxiv.org/abs/2108.13323)"""

    def __init__(
        self,
        student_model,
        teacher_model,
        task_lossfn,
        student_lr=0.1,
        teacher_lr=0.1,
        adaptive_distillation_losses=True,
        adaptive_hidden_losses=True,
        gradient_compression_ratio=1.0,
        user_id=0,
        send_gradient=True,
    ):
        super(FedKDClient, self).__init__(
            student_model, user_id=user_id, lr=student_lr, send_gradient=send_gradient
        )
        self.student_model = student_model
        self.teacher_model = teacher_model
        self.teacher_lr = teacher_lr
        self.task_lossfn = task_lossfn
        self.adaptive_distillation_losses = adaptive_distillation_losses
        self.adaptive_hidden_losses = adaptive_hidden_losses
        self.gradient_compression_ratio = gradient_compression_ratio

        self._is_valid_models()

    def _is_valid_models(self):
        if self.adaptive_hidden_losses:
            if not hasattr(self.teacher_model, "get_hidden_states"):
                raise AttributeError(
                    "If adaptive_hidden_losses=True,\
                 teacher_model must have `get_hidden_states` method"
                )
            if not hasattr(self.student_model, "get_hidden_states"):
                raise AttributeError(
                    "If adaptive_hidden_losses=True,\
                 student_model must have `get_hidden_states` method"
                )

    def loss(self, x, y):
        y_pred_teacher = self.teacher_model(x)
        y_pred_student = self.student_model(x)

        teacher_loss = 0
        student_loss = 0

        # task_losses
        task_loss_teacher = self.task_lossfn(y_pred_teacher, y)
        task_loss_student = self.task_lossfn(y_pred_student, y)
        teacher_loss += task_loss_teacher
        student_loss += task_loss_student

        # adaptive_distillation_losses
        if self.adaptive_distillation_losses:
            adaptive_distillaion_loss_teacher = F.kl_div(
                y_pred_student.softmax(dim=1).log(), y_pred_teacher.softmax(dim=1)
            ) / (task_loss_student + task_loss_teacher)
            adaptive_distillaion_loss_student = F.kl_div(
                y_pred_teacher.softmax(dim=1).log(), y_pred_student.softmax(dim=1)
            ) / (task_loss_student + task_loss_teacher)
            teacher_loss += adaptive_distillaion_loss_teacher
            student_loss += adaptive_distillaion_loss_student

        # adaptove_hidden_losses
        if self.adaptive_hidden_losses:
            adaptive_hidden_losses_student_teacher = 0
            hidden_states_teacher = self.teacher_model.get_hidden_states()
            hidden_states_student = self.student_model.get_hidden_states()

            if type(hidden_states_teacher) != list:
                raise TypeError(
                    "get_hidden_states should return a list of torch.Tensors"
                )
            if type(hidden_states_student) != list:
                raise TypeError(
                    "get_hidden_states should return a list of torch.Tensors"
                )

            for hst, hss in zip(hidden_states_teacher, hidden_states_student):
                adaptive_hidden_losses_student_teacher += F.mse_loss(hst, hss)
            teacher_loss += adaptive_hidden_losses_student_teacher / (
                task_loss_student + task_loss_teacher
            )
            student_loss += adaptive_hidden_losses_student_teacher / (
                task_loss_student + task_loss_teacher
            )

        return teacher_loss, student_loss



"""Subpackage for collaborative learning, where multiple clients trains a single
global model without sharing their local datasets.
"""
from .core import BaseClient, BaseServer  # noqa: F401
from .dsfl import DSFLAPI, DSFLClient, DSFLServer  # noqa : F401
from .fedavg import (  # noqa: F401
    FedAVGAPI,
    FedAVGClient,
    FedAVGServer,
    MPIFedAVGAPI,
    MPIFedAVGClientManager,
    MPIFedAVGServerManager,
)
from .fedgems import FedGEMSAPI, FedGEMSClient, FedGEMSServer  # noqa: F401
from .fedkd import FedKDClient  # noqa :F401
from .fedmd import (  # noqa: F401
    FedMDAPI,
    FedMDClient,
    FedMDServer,
    MPIFedMDAPI,
    MPIFedMDClientManager,
    MPIFedMDServerManager,
)
from .optimizer import AdamFLOptimizer, SGDFLOptimizer  # noqa: F401
from .splitnn import SplitNNAPI, SplitNNClient  # noqa: F401



import numpy as np
from numpy.polynomial import Polynomial


def gen_binary_poly(size):
    """Generates a polynomial with coeffecients in [0, 1]
    Args:
        size: number of coeffcients, size-1 being the degree of the
            polynomial.
    Returns:
        array of coefficients with the coeff[i] being
        the coeff of x ^ i.
    """
    return Polynomial(np.random.randint(0, 2, size, dtype=np.int64))


def gen_uniform_poly(size, modulus):
    """Generates a polynomial with coeffecients being integers in Z_modulus
    Args:
        size: number of coeffcients, size-1 being the degree of the
            polynomial.
    Returns:
        array of coefficients with the coeff[i] being
        the coeff of x ^ i.
    """
    return Polynomial(np.random.randint(0, modulus, size, dtype=np.int64))


def gen_normal_poly(size, mean=0, std=2):
    """Generates a polynomial with coeffecients in a normal distribution
    of mean 0 and a standard deviation of 2, then discretize it.
    Args:
        size: number of coeffcients, size-1 being the degree of the
            polynomial.
    Returns:
        array of coefficients with the coeff[i] being
        the coeff of x ^ i.
    """
    return Polynomial(np.int64(np.random.normal(mean, std, size=size)))



import numpy as np

from aijack_cpp_core import (
    _greedy_search,
    _greedy_search_frac,
    _ternary_search,
    _ternary_search_int,
    culc_tightupperbound_lowerbound_of_rdp_with_theorem6and8_of_zhu_2019,
    culc_upperbound_of_rdp_with_Sampled_Gaussian_Mechanism,
    eps_gaussian,
    eps_laplace,
)

from .rdp import (
    culc_upperbound_of_rdp_with_Sampled_Gaussian_Mechanism as culc_upperbound_of_rdp_with_Sampled_Gaussian_Mechanism_py,
)


class BaseMomentAccountant:
    def __init__(
        self,
        search="ternary",
        order_min=2,
        order_max=64,
        precision=0.5,
        orders=[],
        max_iterations=10000,
    ):
        self.order_min = order_min
        self.order_max = order_max
        self.orders = orders
        self.precision = precision
        self.max_iterations = max_iterations

        self.steps_info = []
        self._culc_bound_of_rdp = None

        if search == "ternary":
            self.search = _ternary_search
        elif search == "ternary_int":
            self.search = _ternary_search_int
        elif search == "greedy":
            self.search = _greedy_search
        elif search == "greedy_frac":
            self.search = _greedy_search_frac

        self._cache = {}

    def _culc_upperbound_of_rdp_onestep(self, alpha, noise_params, sampling_rate):
        key = hash(
            f"{alpha}_{list(noise_params.keys())[0]}_{list(noise_params.values())[0]}_{sampling_rate}"
        )
        if key not in self._cache:
            if sampling_rate == 0:
                result = 0
            elif sampling_rate == 1:
                result = self.eps_func(alpha, noise_params)
            else:
                result = self._culc_bound_of_rdp(
                    alpha, noise_params, sampling_rate, self.eps_func
                )
            self._cache[key] = result
            return result
        else:
            return self._cache[key]

    def _culc_upperbound_of_rdp(self, lam, steps_info):
        rdp = 0.0
        for noise_params, sampling_rate, num_steps in steps_info:
            rdp += num_steps * self._culc_upperbound_of_rdp_onestep(
                lam, noise_params, sampling_rate
            )
        return rdp

    def reset_step_info(self):
        self.steps_info = []

    def add_step_info(self, noise_params, sampling_rate, num_steps):
        self.steps_info.append((noise_params, sampling_rate, num_steps))

    def step(self, noise_params, sampling_rate, num_steps):
        def _step(f):
            def _wrapper(*args, **keywords):
                result = f(*args, **keywords)
                self.add_step_info(self, noise_params, sampling_rate, num_steps)
                return result

            return _wrapper

        return _step

    def get_noise_multiplier(
        self,
        noise_multiplier_key,
        target_epsilon,
        target_delta,
        sampling_rate,
        num_iterations,
        noise_multiplier_min=0,
        noise_multiplier_max=10,
        noise_multiplier_precision=0.01,
    ):
        eps = float("inf")
        while eps > target_epsilon:
            noise_multiplier_max = 2 * noise_multiplier_max
            self.steps_info = [
                (
                    {noise_multiplier_key: noise_multiplier_max},
                    sampling_rate,
                    int(num_iterations / sampling_rate),
                )
            ]
            eps = self.get_epsilon(target_delta)

        while (
            noise_multiplier_max - noise_multiplier_min
        ) > noise_multiplier_precision:
            noise_multiplier = (noise_multiplier_max + noise_multiplier_min) / 2
            self.steps_info = [
                (
                    {noise_multiplier_key: noise_multiplier},
                    sampling_rate,
                    int(num_iterations / sampling_rate),
                )
            ]
            eps = self.get_epsilon(target_delta)

            if eps < target_epsilon:
                noise_multiplier_max = noise_multiplier
            else:
                noise_multiplier_min = noise_multiplier

        return noise_multiplier

    def get_delta(self, epsilon):
        optimal_lam = self.search(
            lambda order: (order - 1)
            * (self._culc_upperbound_of_rdp(order - 1, self.steps_info) - epsilon),
            self.order_min,
            self.order_max,
            self.orders,
            self.precision,
            self.max_iterations,
        )

        min_delta = np.exp(
            (optimal_lam - 1)
            * (self._culc_upperbound_of_rdp(optimal_lam - 1, self.steps_info) - epsilon)
        )

        return min_delta

    def get_epsilon(self, delta):
        # log_inv_delta = math.log(1 / delta)

        def estimate_eps(order):
            return (
                self._culc_upperbound_of_rdp(order, self.steps_info)
                - (np.log(order) + np.log(delta)) / (order - 1)
                + np.log((order - 1) / order)
            )

        optimal_lam = self.search(
            estimate_eps,
            self.order_min,
            self.order_max,
            self.orders,
            self.precision,
            self.max_iterations,
        )

        min_epsilon = estimate_eps(optimal_lam)

        return min_epsilon


class GeneralMomentAccountant(BaseMomentAccountant):
    def __init__(
        self,
        name="SGM",
        search="ternary",
        order_min=2,
        order_max=64,
        precision=0.5,
        orders=[],
        noise_type="Gaussian",
        bound_type="rdp_upperbound_closedformula",
        max_iterations=10000,
        backend="cpp",
    ):
        super().__init__(
            search=search,
            order_min=order_min,
            order_max=order_max,
            precision=precision,
            orders=orders,
            max_iterations=max_iterations,
        )
        self.name = name
        self._set_noise_type(noise_type)
        self._set_upperbound_func(backend, bound_type)

    def _set_noise_type(self, noise_type):
        if noise_type == "Gaussian":
            self.eps_func = eps_gaussian
        elif noise_type == "Laplace":
            self.eps_func = eps_laplace

    def _set_upperbound_func(self, backend, bound_type):
        if backend == "cpp" and bound_type == "rdp_upperbound_closedformula":
            self._culc_bound_of_rdp = (
                culc_upperbound_of_rdp_with_Sampled_Gaussian_Mechanism
            )
        elif backend == "python" and bound_type == "rdp_upperbound_closedformula":
            self._culc_bound_of_rdp = (
                culc_upperbound_of_rdp_with_Sampled_Gaussian_Mechanism_py
            )
        elif backend == "cpp" and bound_type == "rdp_tight_upperbound":
            self._culc_bound_of_rdp = (
                culc_tightupperbound_lowerbound_of_rdp_with_theorem6and8_of_zhu_2019
            )



def l2(fake_gradients, received_gradients, gradient_ignore_pos):
    distance = 0
    for i, (f_g, c_g) in enumerate(zip(fake_gradients, received_gradients)):
        if i not in gradient_ignore_pos:
            distance += ((f_g - c_g) ** 2).sum()
    return distance


def cossim(fake_gradients, received_gradients, gradient_ignore_pos):
    distance = 0
    pnorm_0 = 0
    pnorm_1 = 0
    for i, (f_g, c_g) in enumerate(zip(fake_gradients, received_gradients)):
        if i not in gradient_ignore_pos:
            pnorm_0 = pnorm_0 + f_g.pow(2).sum()
            pnorm_1 = pnorm_1 + c_g.pow(2).sum()
            distance = distance + (f_g * c_g).sum()
    distance = 1 - distance / pnorm_0.sqrt() / pnorm_1.sqrt()
    return distance



import torch
from sklearn.metrics import roc_auc_score

from ...manager import BaseManager


def attach_normattack_to_splitnn(
    cls, attack_criterion, target_client_index=0, device="cpu"
):
    class NormAttackSplitNNWrapper(cls):
        def __init__(self, *args, **kwargs):
            super(NormAttackSplitNNWrapper, self).__init__(*args, **kwargs)
            self.attack_criterion = attack_criterion
            self.target_client_index = target_client_index
            self.device = device

        def extract_intermidiate_gradient(self, outputs):
            self.backward_gradient(outputs.grad)
            return self.clients[self.target_client_index].grad_from_next_client

        def attack(self, dataloader):
            """Culculate leak_auc on the given SplitNN model
            reference: https://arxiv.org/abs/2102.08504
            Args:
                dataloader (torch dataloader): dataloader for evaluation
                criterion: loss function for training
                device: cpu or GPU
            Returns:
                score: culculated leak auc
            """
            epoch_labels = []
            epoch_g_norm = []
            for i, data in enumerate(dataloader, 0):

                inputs, labels = data
                inputs = inputs.to(self.device)
                labels = labels.to(self.device)

                outputs = self(inputs)
                loss = self.attack_criterion(outputs, labels)
                loss.backward()

                grad_from_server = self.extract_intermidiate_gradient(outputs)
                g_norm = grad_from_server.pow(2).sum(dim=1).sqrt()
                epoch_labels.append(labels)
                epoch_g_norm.append(g_norm)

            epoch_labels = torch.cat(epoch_labels)
            epoch_g_norm = torch.cat(epoch_g_norm)
            score = roc_auc_score(epoch_labels, epoch_g_norm.view(-1, 1))
            return score

    return NormAttackSplitNNWrapper


class NormAttackSplitNNManager(BaseManager):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        return attach_normattack_to_splitnn(cls, *self.args, **self.kwargs)



import copy

from ...manager import BaseManager
from ..core import BaseClient
from ..core.utils import GRADIENTS_TAG, PARAMETERS_TAG
from ..optimizer import AdamFLOptimizer, SGDFLOptimizer


class FedAVGClient(BaseClient):
    """Client of FedAVG for single process simulation

    Args:
        model (torch.nn.Module): local model
        user_id (int, optional): if of this client. Defaults to 0.
        lr (float, optional): learning rate. Defaults to 0.1.
        send_gradient (bool, optional): if True, communicate gradient to the server. otherwise, communicates model parameters. Defaults to True.
        optimizer_type_for_global_grad (str, optional): type of optimizer for model update with global gradient. sgd|adam. Defaults to "sgd".
        server_side_update (bool, optional): If True, the global model update is conducted in the server side. Defaults to True.
        optimizer_kwargs_for_global_grad (dict, optional): kwargs for the optimizer for global gradients. Defaults to {}.
        device (str, optional): device type. Defaults to "cpu".
    """

    def __init__(
        self,
        model,
        user_id=0,
        lr=0.1,
        send_gradient=True,
        optimizer_type_for_global_grad="sgd",
        server_side_update=True,
        optimizer_kwargs_for_global_grad={},
        device="cpu",
    ):
        super(FedAVGClient, self).__init__(model, user_id=user_id)
        self.lr = lr
        self.send_gradient = send_gradient
        self.server_side_update = server_side_update
        self.device = device

        if not self.server_side_update:
            self._setup_optimizer_for_global_grad(
                optimizer_type_for_global_grad, **optimizer_kwargs_for_global_grad
            )

        self.prev_parameters = []
        for param in self.model.parameters():
            self.prev_parameters.append(copy.deepcopy(param))

        self.initialized = False

    def _setup_optimizer_for_global_grad(self, optimizer_type, **kwargs):
        if optimizer_type == "sgd":
            self.optimizer_for_gloal_grad = SGDFLOptimizer(
                self.model.parameters(), lr=self.lr, **kwargs
            )
        elif optimizer_type == "adam":
            self.optimizer_for_gloal_grad = AdamFLOptimizer(
                self.model.parameters(), lr=self.lr, **kwargs
            )
        elif optimizer_type == "none":
            self.optimizer_for_gloal_grad = None
        else:
            raise NotImplementedError(
                f"{optimizer_type} is not supported. You can specify `sgd`, `adam`, or `none`."
            )

    def upload(self):
        """Upload the current local model state"""
        if self.send_gradient:
            return self.upload_gradients()
        else:
            return self.upload_parameters()

    def upload_parameters(self):
        """Upload the model parameters"""
        return self.model.state_dict()

    def upload_gradients(self):
        """Upload the local gradients"""
        gradients = []
        for param, prev_param in zip(self.model.parameters(), self.prev_parameters):
            gradients.append((prev_param - param) / self.lr)
        return gradients

    def revert(self):
        """Revert the local model state to the previous global model"""
        for param, prev_param in zip(self.model.parameters(), self.prev_parameters):
            if param is not None:
                param = prev_param

    def download(self, new_global_model):
        """Download the new global model"""
        if self.server_side_update or (not self.initialized):
            # receive the new global model as the model state
            self.model.load_state_dict(new_global_model)
        else:
            # receive the new global model as the global gradients
            self.revert()
            self.optimizer_for_gloal_grad.step(new_global_model)

        if not self.initialized:
            self.initialized = True

        self.prev_parameters = []
        for param in self.model.parameters():
            self.prev_parameters.append(copy.deepcopy(param))

    def local_train(
        self, local_epoch, criterion, trainloader, optimizer, communication_id=0
    ):
        for i in range(local_epoch):
            running_loss = 0.0
            running_data_num = 0
            for _, data in enumerate(trainloader, 0):
                inputs, labels = data
                inputs = inputs.to(self.device)
                labels = labels.to(self.device)

                optimizer.zero_grad()
                self.zero_grad()

                outputs = self(inputs)
                loss = criterion(outputs, labels)

                loss.backward()
                optimizer.step()

                running_loss += loss.item()
                running_data_num += inputs.shape[0]

            print(
                f"communication {communication_id}, epoch {i}: client-{self.user_id+1}",
                running_loss / running_data_num,
            )


def attach_mpi_to_fedavgclient(cls):
    class MPIFedAVGClientWrapper(cls):
        def __init__(self, comm, *args, **kwargs):
            super(MPIFedAVGClientWrapper, self).__init__(*args, **kwargs)
            self.comm = comm

        def action(self):
            self.upload()
            self.model.zero_grad()
            self.download()

        def upload(self):
            self.upload_gradient()

        def upload_gradient(self, destination_id=0):
            self.comm.send(
                super(MPIFedAVGClientWrapper, self).upload_gradients(),
                dest=destination_id,
                tag=GRADIENTS_TAG,
            )

        def download(self):
            super(MPIFedAVGClientWrapper, self).download(
                self.comm.recv(tag=PARAMETERS_TAG)
            )

        def mpi_initialize(self):
            self.download()

    return MPIFedAVGClientWrapper


class MPIFedAVGClientManager(BaseManager):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        return attach_mpi_to_fedavgclient(cls, *self.args, **self.kwargs)



"""Subpackage for poisoning attack, which inserts malicious data to the training dataset,
so that the performance of the trained machine learning model will degregate.
"""
from .history import HistoryAttackClientWrapper  # noqa: F401
from .label_flip import LabelFlipAttackClientManager  # noqa: F401
from .mapf import MAPFClientWrapper  # noqa: F401
from .poison_attack import Poison_attack_sklearn  # noqa: F401



import random

import numpy as np
import torch
from torch.utils.data.dataset import Dataset


def try_gpu(e):
    """Send given tensor to gpu if it is available

    Args:
        e: (torch.Tensor)

    Returns:
        e: (torch.Tensor)
    """
    if torch.cuda.is_available():
        return e.cuda()
    return e


def worker_init_fn(worker_id):
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


class RoundDecimal(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, n_digits):
        ctx.save_for_backward(input)
        ctx.n_digits = n_digits
        return torch.round(input * 10**n_digits) / (10**n_digits)

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = grad_output.clone()
        return torch.round(grad_input * 10**ctx.n_digits) / (10**ctx.n_digits), None


torch_round_x_decimal = RoundDecimal.apply


class NumpyDataset(Dataset):
    """This class allows you to convert numpy.array to torch.Dataset

    Args:
        x (np.array):
        y (np.array):
        transform (torch.transform):

    Attriutes
        x (np.array):
        y (np.array):
        transform (torch.transform):
    """

    def __init__(self, x, y=None, transform=None, return_idx=False):
        self.x = x
        self.y = y
        self.transform = transform
        self.return_idx = return_idx

    def __getitem__(self, index):
        x = self.x[index]
        if self.y is not None:
            y = self.y[index]

        if self.transform is not None:
            x = self.transform(x)

        if not self.return_idx:
            if self.y is not None:
                return x, y
            else:
                return x
        else:
            if self.y is not None:
                return index, x, y
            else:
                return index, x

    def __len__(self):
        """get the number of rows of self.x"""
        return len(self.x)



import math

import numpy as np
from scipy import special


def _log_add(logx, logy):
    """Add two numbers in the log space."""
    a, b = min(logx, logy), max(logx, logy)
    if a == -np.inf:  # adding 0
        return b
    # Use exp(a) + exp(b) = (exp(a - b) + 1) * exp(b)
    return math.log1p(math.exp(a - b)) + b  # log1p(x) = log(x + 1)


def _log_sub(logx, logy):
    """Subtract two numbers in the log space. Answer must be non-negative."""
    if logx < logy:
        raise ValueError("The result of subtraction must be non-negative.")
    if logy == -np.inf:  # subtracting 0
        return logx
    if logx == logy:
        return -np.inf  # 0 is represented as -np.inf in the log space.

    try:
        # Use exp(x) - exp(y) = (exp(x - y) - 1) * exp(y).
        return math.log(math.expm1(logx - logy)) + logy  # expm1(x) = exp(x) - 1
    except OverflowError:
        return logx


def _log_erfc(x):
    """Compute log(erfc(x)) with high accuracy for large x."""
    try:
        return math.log(2) + special.log_ndtr(-x * 2**0.5)
    except NameError:
        # If log_ndtr is not available, approximate as follows:
        r = special.erfc(x)
        if r == 0.0:
            # Using the Laurent series at infinity for the tail of the erfc function:
            #     erfc(x) ~ exp(-x^2-.5/x^2+.625/x^4)/(x*pi^.5)
            # To verify in Mathematica:
            #     Series[Log[Erfc[x]] + Log[x] + Log[Pi]/2 + x^2, {x, Infinity, 6}]
            return (
                -math.log(math.pi) / 2
                - math.log(x)
                - x**2
                - 0.5 * x**-2
                + 0.625 * x**-4
                - 37.0 / 24.0 * x**-6
                + 353.0 / 64.0 * x**-8
            )
        else:
            return math.log(r)



import numpy as np
from numpy.polynomial import Polynomial


def round_coordinates(coordinates):
    """Gives the integral rest."""
    coordinates = coordinates - np.floor(coordinates)
    return coordinates


def coordinate_wise_random_rounding(coordinates):
    """Rounds coordinates randonmly."""
    r = round_coordinates(coordinates)
    f = np.array([np.random.choice([c, c - 1], 1, p=[1 - c, c]) for c in r]).reshape(-1)

    rounded_coordinates = coordinates - f
    rounded_coordinates = [int(coeff) for coeff in rounded_coordinates]
    return rounded_coordinates


def polydiv_coef(p: Polynomial, mod: int) -> Polynomial:
    return Polynomial([c % mod for c in p.coef])



"""Implementation of `Cheng, Sijie, et al. "FedGEMS: Federated Learning of Larger
Server Models via Selective Knowledge Fusion." arXiv preprint arXiv:2110.11027 (2021).`"""
from .api import FedGEMSAPI  # noqa: F401
from .client import FedGEMSClient  # noqa : F401
from .server import FedGEMSServer  # noqa: F401



import torch

from ...utils.metrics import crossentropyloss_between_logits
from ...utils.utils import torch_round_x_decimal
from ..core import BaseClient


class DSFLClient(BaseClient):
    """Client of DS-FL.

    Args:
        model (torch.nn.Module): _description_
        public_dataloader (torch.utils.data.DataLoader): a dataloader of the public dataset.
        output_dim (int, optional): the dimension of the output. Defaults to 1.
        round_decimal (int, optional): number of digits to round up. Defaults to None.
        device (str, optional): device type. Defaults to "cpu".
        user_id (int, optional): id of this client. Defaults to 0.
    """

    def __init__(
        self,
        model,
        public_dataloader,
        output_dim=1,
        round_decimal=None,
        consensus_scale=1.0,
        device="cpu",
        user_id=0,
    ):
        """Init DSFLClient."""
        super().__init__(model, user_id)
        self.public_dataloader = public_dataloader
        self.round_decimal = round_decimal
        self.device = device
        self.global_logit = None
        self.consensus_scale = consensus_scale

        len_public_dataloader = len(self.public_dataloader.dataset)
        self.logit2server = torch.ones((len_public_dataloader, output_dim)).to(
            self.device
        ) * float("inf")

    def upload(self):
        """Upload the output logits on the public dataset to the server.

        Returns:
            torch.Tensor: the output logits of the public dataset.
        """
        for data in self.public_dataloader:
            idx = data[0]
            x = data[1]
            x = x.to(self.device)
            self.logit2server[idx, :] = self(x).detach().softmax(dim=-1)

        if self.round_decimal is None:
            return self.logit2server
        else:
            return torch_round_x_decimal(self.logit2server, self.round_decimal)

    def download(self, global_logit):
        """Download the global logits from the server.

        Args:
            global_logit (torch.Tensor): the global logits from the server
        """
        self.global_logit = global_logit

    def local_train(self, local_epoch, criterion, trainloader, optimizer):

        running_loss = 0.0
        for _ in range(local_epoch):
            for data in trainloader:
                _, x, y = data
                x = x.to(self.device)
                y = y.to(self.device).to(torch.int64)

                optimizer.zero_grad()
                loss = criterion(self(x), y)
                loss.backward()
                optimizer.step()

                running_loss += loss.item()

        return running_loss

    def approach_consensus(self, consensus_optimizer):
        """Train the own local model to minimize the distance between the global logits and
        the output logits of the local model on the public dataset.

        Args:
            consensus_optimizer (torch.optim.Optimizer): an optimizer to train the local model.

        Returns:
            float: averaged loss.
        """
        running_loss = 0
        for global_data in self.public_dataloader:
            idx = global_data[0]
            x = global_data[1].to(self.device)
            y_global = self.global_logit[idx, :].to(self.device).detach()
            consensus_optimizer.zero_grad()
            y_local = self(x)
            loss_consensus = self.consensus_scale * crossentropyloss_between_logits(
                y_local, y_global
            )
            loss_consensus.backward()
            consensus_optimizer.step()
            running_loss += loss_consensus.item()
        running_loss /= len(self.public_dataloader)
        return running_loss



import copy

import torch

from ..core.api import BaseFedAPI, BaseFLKnowledgeDistillationAPI


class FedMDAPI(BaseFLKnowledgeDistillationAPI):
    """Implementation of `Fedmd: Heterogenous federated learning via model distillation`"""

    def __init__(
        self,
        server,
        clients,
        public_dataloader,
        local_dataloaders,
        criterion,
        client_optimizers,
        validation_dataloader=None,
        server_optimizer=None,
        num_communication=1,
        device="cpu",
        consensus_epoch=1,
        revisit_epoch=1,
        transfer_epoch_public=1,
        transfer_epoch_private=1,
        server_training_epoch=1,
        custom_action=lambda x: x,
    ):
        super().__init__(
            server,
            clients,
            public_dataloader,
            local_dataloaders,
            validation_dataloader,
            criterion,
            num_communication,
            device,
        )
        self.client_optimizers = client_optimizers
        self.server_optimizer = server_optimizer
        self.consensus_epoch = consensus_epoch
        self.revisit_epoch = revisit_epoch
        self.transfer_epoch_public = transfer_epoch_public
        self.transfer_epoch_private = transfer_epoch_private
        self.server_training_epoch = server_training_epoch
        self.custom_action = custom_action
        self.epoch = 0

    def train_server(self):
        if self.server_optimizer is None:
            return 0.0

        running_loss = 0.0
        for data in self.public_dataloader:
            _, x, y = data
            x = x.to(self.device)
            y = y.to(self.device).to(torch.int64)

            self.server_optimizer.zero_grad()
            loss = self.criterion(self.server(x), y)
            loss.backward()
            self.server_optimizer.step()

            running_loss += loss.item()

        running_loss /= len(self.public_dataloader)

        return running_loss

    def run(self):
        logging = {
            "loss_client_local_dataset_transfer": [],
            "loss_client_public_dataset_transfer": [],
            "loss_client_consensus": [],
            "loss_client_revisit": [],
            "loss_server_public_dataset": [],
            "acc_local": [],
            "acc_pub": [],
            "acc_val": [],
        }

        # Transfer
        for i in range(1, self.transfer_epoch_public + 1):
            loss_public = self.train_client(public=True)
            print(f"epoch {i} (public - pretrain): {loss_public}")
            logging["loss_client_public_dataset_transfer"].append(loss_public)

        if self.validation_dataloader is not None:
            acc_val = self.score(
                self.validation_dataloader, self.server_optimizer is None
            )
            print("acc on validation dataset: ", acc_val)
            logging["acc_val"].append(copy.deepcopy(acc_val))

        for i in range(1, self.transfer_epoch_private + 1):
            loss_local = self.train_client(public=False)
            print(f"epoch {i} (local - pretrain): {loss_local}")
            logging["loss_client_local_dataset_transfer"].append(loss_local)

        if self.validation_dataloader is not None:
            acc_val = self.score(
                self.validation_dataloader, self.server_optimizer is None
            )
            print("acc on validation dataset: ", acc_val)
            logging["acc_val"].append(copy.deepcopy(acc_val))

        for i in range(1, self.num_communication + 1):

            self.epoch = i

            self.server.action()

            # Digest
            temp_consensus_loss = []
            if len(self.clients) > 1:
                for j, client in enumerate(self.clients):
                    for _ in range(self.consensus_epoch):
                        consensus_loss = client.approach_consensus(
                            self.client_optimizers[j]
                        )
                    print(f"epoch {i}, client {j}: {consensus_loss}")
                    temp_consensus_loss.append(consensus_loss)
            logging["loss_client_consensus"].append(temp_consensus_loss)

            # Revisit
            for _ in range(self.revisit_epoch):
                loss_local_revisit = self.train_client(public=False)
            logging["loss_client_revisit"].append(loss_local_revisit)

            # Train a server-side model if it exists (different from the original paper)
            for _ in range(self.server_training_epoch):
                loss_server_public = self.train_server()
            logging["loss_server_public_dataset"].append(loss_server_public)

            acc_on_local_dataset = self.local_score()
            print(f"epoch={i} acc on local datasets: ", acc_on_local_dataset)
            logging["acc_local"].append(acc_on_local_dataset)
            acc_pub = self.score(self.public_dataloader, self.server_optimizer is None)
            print(f"epoch={i} acc on public dataset: ", acc_pub)
            logging["acc_pub"].append(copy.deepcopy(acc_pub))
            # evaluation
            if self.validation_dataloader is not None:
                acc_val = self.score(
                    self.validation_dataloader, self.server_optimizer is None
                )
                print(f"epoch={i} acc on validation dataset: ", acc_val)
                logging["acc_val"].append(copy.deepcopy(acc_val))

            self.custom_action(self)

        return logging


class MPIFedMDAPI(BaseFedAPI):
    def __init__(
        self,
        comm,
        party,
        is_server,
        criterion,
        local_optimizer=None,
        local_dataloader=None,
        public_dataloader=None,
        num_communication=1,
        local_epoch=1,
        consensus_epoch=1,
        revisit_epoch=1,
        transfer_epoch_public=1,
        transfer_epoch_private=1,
        custom_action=lambda x: x,
        device="cpu",
    ):
        self.comm = comm
        self.party = party
        self.is_server = is_server
        self.criterion = criterion
        self.local_optimizer = local_optimizer
        self.local_dataloader = local_dataloader
        self.public_dataloader = public_dataloader
        self.num_communication = num_communication
        self.local_epoch = local_epoch
        self.consensus_epoch = consensus_epoch
        self.revisit_epoch = revisit_epoch
        self.transfer_epoch_public = transfer_epoch_public
        self.transfer_epoch_private = transfer_epoch_private
        self.custom_action = custom_action
        self.device = device

    def run(self):
        for _ in range(self.num_communication):

            # Transfer phase
            if not self.is_server:
                self.local_train(epoch=self.transfer_epoch_public, public=True)
                self.local_train(epoch=self.transfer_epoch_private, public=False)

            # Updata global logits
            self.party.action()
            self.comm.Barrier()

            # Digest phase
            if not self.is_server:
                for _ in range(1, self.consensus_epoch):
                    self.party.approach_consensus(self.local_optimizer)

            # Revisit phase
            if not self.is_server:
                for _ in range(self.revisit_epoch):
                    self.local_train(public=False)

            self.custom_action(self)
            self.comm.Barrier()

    def local_train(self, epoch=1, public=True):
        trainloader = self.public_dataloader if public else self.local_dataloader
        self.party.local_train(epoch, self.criterion, trainloader, self.local_optimizer)



from aijack_cpp_core import (  # noqa: F401
    PaillierCipherText,
    PaillierKeyGenerator,
    PaillierPublicKey,
    PaillierSecretKey,
)

from .fed_wrapper import PaillierGradientClientManager  # noqa: F401
from .torch_wrapper import PaillierTensor  # noqa: F401



import copy

import torch

from ..core.api import BaseFLKnowledgeDistillationAPI


class FedGEMSAPI(BaseFLKnowledgeDistillationAPI):
    """API of FedGEMSAPI.

    Args:
        server (FedGEMSServer): a server.
        clients (List[FedGEMSClient]): a list of clients.
        public_dataloader (torch.utils.data.DataLoader): a dataloader of the public dataset.
        local_dataloaders (List[torch.utils.data.DataLoader]): a list of dataloaders of
                                                               the local datasets.
        validation_dataloader (torch.utils.data.DataLoader): a dataloader of the validation dataset.
        criterion (function)): a loss function
        server_optimizer (torch.optim.Optimizer): an optimizer for the global model
        client_optimizers (List[torch.optim.Optimizer]): a list of optimizers for the local models
        num_communication (int, optional): the number of communications. Defaults to 10.
        epoch_client_on_localdataset (int, optional): the number of epochs of client-side
                                                      training on the private datasets.
                                                      Defaults to 10.
        epoch_client_on_publicdataset (int, optional): the number of epochs of client-side
                                                       training on the public datasets.
                                                       Defaults to 10.
        epoch_server_on_publicdataset (int, optional): the number of epochs of server-side training
                                                       on the public dataset. Defaults to 10.
        device (str, optional): device type. Defaults to "cpu".
        custom_action (function, optional): custom function which this api calls at
                                            the end of every communication.
                                            Defaults to lambda x:x.
    """

    def __init__(
        self,
        server,
        clients,
        public_dataloader,
        local_dataloaders,
        criterion,
        server_optimizer,
        client_optimizers,
        validation_dataloader=None,
        num_communication=10,
        epoch_client_on_localdataset=10,
        epoch_client_on_publicdataset=10,
        epoch_server_on_publicdataset=10,
        device="cpu",
        custom_action=lambda x: x,
    ):
        super().__init__(
            server,
            clients,
            public_dataloader,
            local_dataloaders,
            validation_dataloader,
            criterion,
            num_communication=num_communication,
            device=device,
        )
        self.server_optimizer = server_optimizer
        self.client_optimizers = client_optimizers

        self.epoch_client_on_localdataset = epoch_client_on_localdataset
        self.epoch_client_on_publicdataset = epoch_client_on_publicdataset
        self.epoch_server_on_publicdataset = epoch_server_on_publicdataset

        self.custom_action = custom_action
        self.epoch = 0

    def train_client_on_public_dataset(self):
        """Train clients on the public dataset.

        Returns:
            List[float]: a list of average loss of each client.
        """
        loss_on_public_dataset = []
        for client_idx in range(self.client_num):
            client = self.clients[client_idx]
            optimizer = self.client_optimizers[client_idx]

            running_loss = 0.0
            for data in self.public_dataloader:
                idx, inputs, labels = data
                inputs = inputs.to(self.device)
                labels = labels.to(self.device).to(torch.int64)

                optimizer.zero_grad()
                y_preds = client(inputs)
                loss = client.culc_loss_on_public_dataset(idx, y_preds, labels)
                loss.backward()
                optimizer.step()

                running_loss += loss.item()

            loss_on_public_dataset.append(
                copy.deepcopy(running_loss / len(self.public_dataloader))
            )

        return loss_on_public_dataset

    def train_server_on_public_dataset(self):
        """Train the global model on the public dataset.

        Returns:
            float: average loss
        """
        server_running_loss = 0
        for data in self.public_dataloader:
            idx, x, y = data
            x = x.to(self.device)
            y = y.to(self.device).to(torch.int64)

            self.server_optimizer.zero_grad()
            server_loss = self.server.self_evaluation_on_the_public_dataset(idx, x, y)

            total_loss = server_loss  # + 5*attack_loss
            total_loss.backward()
            self.server_optimizer.step()
            self.server.update(idx, x)

            server_running_loss += server_loss.item() / len(self.public_dataloader)

        self.server.action()
        return server_running_loss

    def run(self):
        logging = {
            "loss_client_local_dataset": [],
            "loss_server_public_dataset": [],
            "loss_client_public_dataset": [],
            "acc_local": [],
            "acc_pub": [],
            "acc_val": [],
        }

        # train FedGEMS
        for epoch in range(1, self.num_communication + 1):

            self.epoch = epoch

            for _ in range(self.epoch_client_on_localdataset):
                loss_client_local_dataset = self.train_client(public=False)
            for _ in range(self.epoch_server_on_publicdataset):
                loss_server_public_dataset = self.train_server_on_public_dataset()
            for _ in range(self.epoch_client_on_publicdataset):
                loss_client_public_dataset = self.train_client_on_public_dataset()

            print(
                f"epoch={epoch} loss_client_local_dataset: ", loss_client_local_dataset
            )
            logging["loss_client_local_dataset"].append(
                copy.deepcopy(loss_client_local_dataset)
            )
            print(
                f"epoch={epoch} loss_server_public_dataset: ",
                loss_server_public_dataset,
            )
            logging["loss_server_public_dataset"].append(
                copy.deepcopy(loss_server_public_dataset)
            )
            print(
                f"epoch={epoch} loss_client_public_dataset: ",
                loss_client_public_dataset,
            )
            logging["loss_client_public_dataset"].append(
                copy.deepcopy(loss_client_public_dataset)
            )

            acc_on_local_dataset = self.local_score()
            print(f"epoch={epoch} acc on local datasets: ", acc_on_local_dataset)
            logging["acc_local"].append(acc_on_local_dataset)
            acc_pub = self.score(self.public_dataloader)
            print(f"epoch={epoch} acc on public dataset: ", acc_pub)
            logging["acc_pub"].append(copy.deepcopy(acc_pub))
            # evaluation
            if self.validation_dataloader is not None:
                acc_val = self.score(self.validation_dataloader)
                print(f"epoch={epoch} acc on validation dataset: ", acc_val)
                logging["acc_val"].append(copy.deepcopy(acc_val))

            self.custom_action(self)

        return logging



import random

import numpy as np
import torch
import torchvision
import torchvision.transforms as transforms

from .utils import NumpyDataset


def prepareFederatedMNISTDataloaders(
    client_num=2,
    local_label_num=2,
    local_data_num=20,
    batch_size=1,
    test_batch_size=16,
    path="MNIST/.",
    download=True,
    transform=transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,))]
    ),
    seed=0,
    return_idx=False,
):
    np.random.seed(seed)
    random.seed(seed)

    at_t_dataset_train = torchvision.datasets.MNIST(
        root=path, train=True, download=download
    )
    at_t_dataset_test = torchvision.datasets.MNIST(
        root=path, train=False, download=download
    )

    X = at_t_dataset_train.train_data.numpy()
    y = at_t_dataset_train.train_labels.numpy()

    test_set = NumpyDataset(
        at_t_dataset_test.test_data.numpy(),
        at_t_dataset_test.test_labels.numpy(),
        transform=transform,
        return_idx=return_idx,
    )
    testloader = torch.utils.data.DataLoader(
        test_set, batch_size=test_batch_size, shuffle=True, num_workers=0
    )

    trainloaders = []
    train_sizes = []
    idx_used = []
    for c in range(client_num):
        assigned_labels = random.sample(range(10), local_label_num)
        print(f"the labels that client_id={c} has are: ", assigned_labels)
        idx = np.concatenate([np.where(y == al)[0] for al in assigned_labels])
        assigned_idx = random.sample(list(set(idx) - set(idx_used)), local_data_num)

        temp_trainset = NumpyDataset(
            X[assigned_idx], y[assigned_idx], transform=transform, return_idx=return_idx
        )
        temp_trainloader = torch.utils.data.DataLoader(
            temp_trainset, batch_size=batch_size, shuffle=True, num_workers=0
        )
        trainloaders.append(temp_trainloader)
        train_sizes.append(len(temp_trainset))

        idx_used += assigned_idx

    return X, y, trainloaders, testloader, train_sizes, idx_used



from numpy.polynomial import Polynomial

from .ciphertext import CKKSCiphertext
from .plaintext import CKKSPlaintext
from .random import gen_binary_poly, gen_normal_poly, gen_uniform_poly
from .utils import polydiv_coef


class CKKSEncrypter:
    def __init__(self, encoder, q):
        self.N = encoder.N
        self.q = q
        self.encoder = encoder
        self.scale = encoder.scale

        self.poly_mod = Polynomial(
            [1 if i == 0 or i == self.N else 0 for i in range(self.N + 1)]
        )

        self.pk = None
        self.sk = None

    def set_pk(self, pk):
        self.pk = pk

    def set_sk(self, sk):
        self.sk = sk

    def keygen(self, size):
        """Generate a public and secret keys
        Args:
            size: size of the polynoms for the public and secret keys.
            q: coefficient modulus.
            poly_mod: polynomial modulus.
        Returns:
            Public and secret key.
        """
        self.sk = gen_binary_poly(size)
        a = gen_uniform_poly(size, self.q)
        e = gen_normal_poly(size, std=1)
        b = (
            polydiv_coef(polydiv_coef(-a * self.sk, self.q) % self.poly_mod - e, self.q)
            % self.poly_mod
        )
        self.pk = (b, a)
        return self.pk, self.sk

    def encrypt_from_plaintext(self, pt):
        """Encrypt an integer.
        Args:
            pk: public-key.
            pt: plaintext to be encrypted.
        Returns:
            Tuple representing a ciphertext.
        """
        return CKKSCiphertext(
            polydiv_coef(pt.p + self.pk[0], self.q) % self.poly_mod,
            self.pk[1],
            self.N,
            self.q,
            self.scale,
        )

    def decrypt_to_plaintext(self, ct):
        return CKKSPlaintext(
            polydiv_coef(
                ct.c0 + polydiv_coef(ct.c1 * self.sk, self.q) % self.poly_mod, self.q
            )
            % self.poly_mod,
            self.N,
            self.scale,
        )

    def encrypt(self, z):
        return self.encrypt_from_plaintext(self.encoder.encode(z))

    def decrypt(self, ct):
        return self.encoder.decode(self.decrypt_to_plaintext(ct))



import copy

import numpy as np
import sklearn
from tqdm import tqdm

from ..base_attack import BaseAttacker


class Poison_attack_sklearn(BaseAttacker):
    """implementation of poison attack for sklearn binary classifier
        reference https://arxiv.org/abs/1206.6389

    Args:
        target_model: sklean classifier
        X_train: training data for target_model
        y_train: training label for target_model
        t: step size

    Attributes:
        target_model: sklean classifier
        X_train:
        y_train:
        t: step size
        kernel
        delta_kernel
    """

    def __init__(self, target_model, X_train, y_train, t=0.5):
        super().__init__(target_model)

        self.X_train = X_train
        self.y_train = y_train
        self.t = t

        self.kernel = None
        self.delta_kernel = None

        _ = self._detect_type_of_classifier()

    def _detect_type_of_classifier(self):
        """detect the type of classifier and prepare proper settings

        Returns:
            return true if no error occurs

        Raises:
            ValueError: if given kernel type is not supported.
        """
        target_type = type(self.target_model)

        if target_type == sklearn.svm._classes.SVC:
            params = self.target_model.get_params()
            kernel_type = params["kernel"]
            if kernel_type == "linear":
                self.kernel = lambda xa, xb: xa.dot(xb.T)
                self.delta_kernel = lambda xi, xc: self.t * xi
            else:
                raise ValueError(f"kernel type {kernel_type} is not supported")

        else:
            raise ValueError(f"target type {target_type} is not supported")

        return True

    def _delta_q(self, xi, xc, yi, yc):
        """Culculate deviation of q
           Q = yy.T * K denotes the label - annotated version of K,
           and α denotes the SVM’s dual variables corresponding
           to each training point.

        Args:
            xi: intermidiate results of the generation of adversarial example
            xc: initial attack point
            yi: the labels of intermidiate results of the generation
                of adversarial example
            yc: true label of initial attack point

        Returns:
            dq:
        """
        d = xi.shape[1]
        yy = np.array([(yi * yc)] * d).T
        dq = yy * (self.delta_kernel(xi, xc))
        return dq

    def attack(self, xc, yc, X_valid, y_valid, num_iterations=200):
        """Create an adversarial example for poison attack

        Args:
            xc: initial attack point
            yc: true label of initial attack point
            X_valid: validation data for target_model
            y_valid: validation label for target_model
            num_iterations: (default = 200)

        Returns:
            xc: created adversarial example
            log: log of score of target_model under attack
        """
        # flip the class label
        yc *= -1
        log = []
        X_train_poisoned = copy.copy(self.X_train)
        y_train_poisoned = copy.copy(self.y_train)

        # best_score = float("inf")
        # best_xc = None
        # best_itr = None

        for i in tqdm(range(num_iterations)):

            target_model_ = copy.copy(self.target_model)
            target_model_.__init__()
            target_model_.set_params(**self.target_model.get_params())
            # target_model_ = sklearn.svm.SVC(kernel="linear", C=1)

            # add poinsoned data
            target_model_.fit(
                np.concatenate([X_train_poisoned, xc.reshape(1, -1)]),
                np.concatenate([y_train_poisoned, [yc]]),
            )

            score_temp = target_model_.score(X_valid, y_valid)
            log.append(score_temp)
            # if score_temp < best_score:
            #    best_score = score_temp
            #    best_xc = xc
            #    best_itr = i

            # ------------------------ #
            xs = target_model_.support_vectors_
            ys = np.concatenate([y_train_poisoned, [yc]])[target_model_.support_]

            Qks = y_valid.reshape(-1, 1).dot(ys.reshape(-1, 1).T) * self.kernel(
                X_valid, xs
            )
            Qss_inv = np.linalg.inv(self.kernel(xs, xs))
            v = Qss_inv.dot(ys)
            zeta = ys.T.dot(v)
            Mk = (-1 / zeta) * (
                (Qks).dot(zeta * Qss_inv - v.dot(v.T))
                + y_valid.reshape(-1, 1).dot(v.reshape(1, -1))
            )

            delta_Qsc = self._delta_q(xs, xc.reshape, ys, yc)
            delta_Qkc = self._delta_q(X_valid, xc.reshape(1, -1), y_valid, yc)

            # α denotes the SVM’s dual variables corresponding to each
            # training point
            alpha = target_model_.decision_function([xc])

            # the desired gradient used for optimizing our attack:
            delta_L = np.sum(((Mk.dot(delta_Qsc) + delta_Qkc) * alpha), axis=0)

            # u is a norm-1 vector representing the attack direction,
            u = delta_L / np.sqrt(np.sum((delta_L**2)))

            # the attack point
            xc += self.t * u

        # print(f"initial score is {log[0]}")
        # print(f"best score is {best_score} in iteration {best_itr}")

        return xc, log



from .api import FedMDAPI, MPIFedMDAPI  # noqa : F401
from .client import FedMDClient, MPIFedMDClientManager  # noqa : F401
from .nfdp import (  # noqa : F401
    get_delta_of_fedmd_nfdp,
    get_epsilon_of_fedmd_nfdp,
    get_k_of_fedmd_nfdp,
)
from .server import FedMDServer, MPIFedMDServerManager  # noqa : F401



from numpy.polynomial import Polynomial


class CKKSPlaintext:
    def __init__(self, p: Polynomial, N: int, scale: int):
        self.p = p
        self.N = N
        self.scale = scale

        self.poly_modulo = Polynomial(
            [1 if i == 0 or i == N else 0 for i in range(N + 1)]
        )

    def __add__(self, other):
        if type(other) in [int, float]:
            return CKKSPlaintext(self.p + other, self.N, self.scale)
        elif type(other) == CKKSPlaintext:
            return CKKSPlaintext(self.p + other.p, self.N, self.scale)
        else:
            raise TypeError("error!")

    def __mul__(self, other):
        if type(other) in [int, float]:
            return CKKSPlaintext(self.p * other, self.N, self.scale)
        elif type(other) == CKKSPlaintext:
            return CKKSPlaintext(
                ((self.p * other.p) / self.scale) % self.poly_modulo, self.N, self.scale
            )
        else:
            raise TypeError("error!")



import torch
import torch.nn as nn
import torch.nn.functional as F


class Purifier_Cifar10(nn.Module):
    """autoencoder for purification on Cifar10
    reference https://arxiv.org/abs/2005.03915
    """

    def __init__(self):
        super(Purifier_Cifar10, self).__init__()
        self.L1 = nn.Linear(10, 7)
        self.bn1 = nn.BatchNorm1d(7)
        self.L2 = nn.Linear(7, 4)
        self.bn2 = nn.BatchNorm1d(4)
        self.L3 = nn.Linear(4, 7)
        self.bn3 = nn.BatchNorm1d(7)
        self.L4 = nn.Linear(7, 10)

    def forward(self, x):
        # 10 -> 7
        x = self.L1(x)
        x = self.bn1(x)
        x = F.relu(x)
        # 7 -> 4
        x = self.L2(x)
        x = self.bn2(x)
        x = F.relu(x)
        # 4 -> 7
        x = self.L3(x)
        x = self.bn3(x)
        x = F.relu(x)
        # 7 -> 10
        x = self.L4(x)

        return x


def PurifierLoss(
    prediction,
    pred_purified,
    lam=0.2,
    purifier_criterion=nn.MSELoss(),
    accuracy_criterion=nn.CrossEntropyLoss(),
):
    """basic loss function for purification
       reference https://arxiv.org/abs/2005.03915

       train purifier G against target model F to minimize the
       following objective function

            L(G) = E[R(G(F(x)), F(x)) + λC(G(F(x), argmax F(x)))]

       R is a reconstruction loss function
       C is a cross entropy loss function
       λ controls the balance of the two loss functions

    Args:
        prediction: predicted value of target model
        pred_purified: purified predicted value of target model
        lam: controls the balance of the following two functions
        purifier_criterion: loss function to reshapre confidense score
        accuracy_criterion: loss function to preserve the accuracy (C)

    Return:
        loss_purifier: weighted average of the two loss function
    """

    loss_1 = purifier_criterion(pred_purified, prediction)
    loss_2 = accuracy_criterion(pred_purified, torch.argmax(prediction, axis=1))
    loss_purifier = loss_1 + lam * loss_2

    return loss_purifier



import copy

from ..core.api import BaseFLKnowledgeDistillationAPI


class DSFLAPI(BaseFLKnowledgeDistillationAPI):
    """API of DS-FL

    Args:
        server (DSFLServer): an instance of DSFLServer
        clients (List[DSFLClient]): a list of instances of DSFLClient
        public_dataloader (torch.DataLoader): a dataloader of public dataset
        local_dataloaders (List[torch.DataLoader]): a list of dataloaders of private dataests
        validation_dataloader (torch.DataLoader): a dataloader of validation dataset
        criterion (function): a loss function
        num_communication (int): number of communication
        device (str): device type
        server_optimizer (torch.Optimizer): a optimizer for the global model
        client_optimizers ([torch.Optimizer]): a list of optimizers for the local models
        epoch_local_training (int, optional): number of epochs of local training. Defaults to 1.
        epoch_global_distillation (int, optional): number of epochs of global distillation. Defaults to 1.
        epoch_local_distillation (int, optional): number of epochs of local distillation. Defaults to 1.
    """

    def __init__(
        self,
        server,
        clients,
        public_dataloader,
        local_dataloaders,
        criterion,
        num_communication,
        device,
        server_optimizer,
        client_optimizers,
        validation_dataloader=None,
        epoch_local_training=1,
        epoch_global_distillation=1,
        epoch_local_distillation=1,
        custom_action=lambda x: x,
    ):
        """Init DSFLAPI"""
        super().__init__(
            server,
            clients,
            public_dataloader,
            local_dataloaders,
            validation_dataloader,
            criterion,
            num_communication,
            device,
        )
        self.server_optimizer = server_optimizer
        self.client_optimizers = client_optimizers
        self.epoch_local_training = epoch_local_training
        self.epoch_global_distillation = epoch_global_distillation
        self.epoch_local_distillation = epoch_local_distillation

        self.custom_action = custom_action
        self.epoch = 0

    def run(self):
        logging = {
            "loss_local": [],
            "loss_client_consensus": [],
            "loss_server_consensus": [],
            "acc_local": [],
            "acc_val": [],
        }
        for i in range(1, self.num_communication + 1):

            self.epoch = i

            for _ in range(self.epoch_local_training):
                loss_local = self.train_client(public=False)
            logging["loss_local"].append(loss_local)

            self.server.action()

            acc_on_local_dataset = self.local_score()
            print(f"epoch={i} acc on local datasets: ", acc_on_local_dataset)
            logging["acc_local"].append(acc_on_local_dataset)

            # distillation
            for _ in range(self.epoch_global_distillation):
                loss_global = self.server.update_globalmodel(self.server_optimizer)
            logging["loss_server_consensus"].append(loss_global)

            self.custom_action(self)

            temp_consensus_loss = []
            if len(self.clients) > 1:
                for j, client in enumerate(self.clients):
                    for _ in range(self.epoch_local_distillation):
                        consensus_loss = client.approach_consensus(
                            self.client_optimizers[j]
                        )
                    temp_consensus_loss.append(consensus_loss)
            logging["loss_client_consensus"].append(temp_consensus_loss)

            print(f"epoch {i}: loss_local", loss_local)
            print(f"epoch {i}: loss_client_consensus", temp_consensus_loss)
            print(f"epoch {i}: loss_server_consensus", loss_global)

            # validation
            if self.validation_dataloader is not None:
                acc_val = self.score(self.validation_dataloader)
                print(f"epoch={i} acc on validation dataset: ", acc_val)
                logging["acc_val"].append(copy.deepcopy(acc_val))

        return logging



import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim

from ...utils import NumpyDataset, try_gpu
from ..base_attack import BaseAttacker


class ShadowModel:
    """Train shadow models for membership inference
       reference https://arxiv.org/abs/1610.05820

    Args
        models : torch models for shadow
        shadow_dataset_suze (int) : size of dataset for shadow models
        shadow_transfomrm (torch.transform) : transform
        seed (int) : random seed

    Attriutes
        models : torch models for shadow
        shadow_dataset_suze (int) : size of dataset for shadow models
        shadow_transfomrm (torch.transform) : transform
        seed (int) : random seed
        num_models(int):
        trainloaders
        testloaders

    """

    def __init__(
        self,
        models,
        shadow_dataset_size,
        shadow_transform=None,
        seed=42,
    ):

        self.models = models
        self.shadow_dataset_size = shadow_dataset_size
        self.shadow_transform = shadow_transform
        self.seed = seed

        self.num_models = len(models)
        self.trainloaders = []
        self.testloaders = []

        # initialize random state
        self._reset_random_state()

    def _reset_random_state(self):
        """initialize random state"""
        self._prng = np.random.RandomState(self.seed)

    def fit_transform(self, X, y, num_itr=100):
        """train shadow models and get prediction of shadow models
           and membership label of each prediction for each class

        Args:
            X (np.array): target data
            y (np.array): target label
            num_itr (int): number of iteration for training

        Returns:
            result_dict (dict) : key is each class
                                 value is (shadow_data, shadow_label)
        """
        self._fit(X, y, num_itr=num_itr)
        (
            shadow_in_data,
            shadow_out_data,
            in_original_labels,
            out_original_labels,
        ) = self._transform()

        shadow_in_label = torch.ones_like(in_original_labels)
        shadow_out_label = torch.zeros_like(out_original_labels)
        shadow_data = torch.cat([shadow_in_data, shadow_out_data])
        shadow_label = torch.cat([shadow_in_label, shadow_out_label])
        original_labels = torch.cat([in_original_labels, out_original_labels])

        result_dict = {}
        unique_labels = torch.unique(original_labels)
        for label in unique_labels:
            local_idx = torch.where(original_labels == label)
            result_dict[label.item()] = (
                shadow_data[local_idx],
                shadow_label[local_idx],
            )

        return result_dict

    def _fit(self, X, y, num_itr=100):
        """train shadow models on given data

        Args:
            X (np.array): training data for shadow models
            y (np.array): training label for shadow models
            num_itr (int): number of iteration for training
        """

        indices = np.arange(X.shape[0])

        for model_idx in range(self.num_models):
            model = self.models[model_idx]

            shadow_indices = self._prng.choice(
                indices, 2 * self.shadow_dataset_size, replace=False
            )

            train_indices = shadow_indices[: self.shadow_dataset_size]
            test_indices = shadow_indices[self.shadow_dataset_size :]

            X_train, y_train = X[train_indices], y[train_indices]
            X_test, y_test = X[test_indices], y[test_indices]

            trainset = NumpyDataset(X_train, y_train, transform=self.shadow_transform)
            trainloader = torch.utils.data.DataLoader(
                trainset, batch_size=4, shuffle=True, num_workers=2
            )
            testset = NumpyDataset(X_test, y_test, transform=self.shadow_transform)
            testloader = torch.utils.data.DataLoader(
                testset, batch_size=4, shuffle=True, num_workers=2
            )

            self.trainloaders.append(trainloader)
            self.testloaders.append(testloader)

            # training
            # TODO : allow users to use specified loss function and optimizer
            criterion = nn.CrossEntropyLoss()
            optimizer = optim.SGD(model.parameters(), lr=0.001, momentum=0.9)

            for epoch in range(num_itr):
                running_loss = 0.0
                for i, data in enumerate(trainloader, 0):
                    inputs, labels = data
                    inputs = try_gpu(inputs)
                    labels = try_gpu(labels)
                    optimizer.zero_grad()
                    outputs = model(inputs)

                    loss = criterion(outputs, labels)
                    loss.backward()
                    optimizer.step()

                    running_loss += loss.item()
                    if i % 2000 == 1999:
                        print(
                            "[%d, %5d] loss: %.3f"
                            % (epoch + 1, i + 1, running_loss / 2000)
                        )
                        running_loss = 0.0
            print("Finished Training")

            self.models[model_idx] = model

    def _transform(self):
        """get prediction and its membership label per each class
           from shadow models

        Returns:
            shadow_in_data (torch.Tensor): prediction from shadow model
                                           on in_data
                                           (in_data means training data of
                                            each shadow model)
            shadow_out_data (torch.Tensor): prediction from shadow model
                                            on out_data
                                           (out_data means the data which
                                            each shadow model has not seen)
            in_original_labels (torch.Tensor): membership labels for
                                               prediciton on in_data
                                               (they should be all positive)
            out_original_labels (torch.Tensor): membership labels for
                                                prediciton on out__data
                                               (they should be all negative)
        """

        shadow_in_data = []
        shadow_out_data = []
        in_original_labels = []
        out_original_labels = []

        for model_idx in range(self.num_models):
            model = self.models[model_idx]
            trainloader = self.trainloaders[model_idx]
            testloader = self.testloaders[model_idx]

            # shadow-in
            train_preds = []
            train_label = []
            with torch.no_grad():
                for data in trainloader:
                    inputs, labels = data
                    inputs = try_gpu(inputs)
                    labels = try_gpu(labels)
                    outputs = model(inputs)
                    train_preds.append(outputs)
                    train_label.append(labels)
            train_preds = torch.cat(train_preds)
            shadow_in_data.append(train_preds)
            train_label = torch.cat(train_label)
            in_original_labels.append(train_label)

            # shadow-out
            test_preds = []
            test_label = []
            with torch.no_grad():
                for data in testloader:
                    inputs, labels = data
                    inputs = try_gpu(inputs)
                    labels = try_gpu(labels)
                    outputs = model(inputs)
                    test_preds.append(outputs)
                    test_label.append(labels)
            test_preds = torch.cat(test_preds)
            shadow_out_data.append(test_preds)
            test_label = torch.cat(test_label)
            out_original_labels.append(test_label)

        shadow_in_data = torch.cat(shadow_in_data)
        shadow_out_data = torch.cat(shadow_out_data)
        in_original_labels = torch.cat(in_original_labels)
        out_original_labels = torch.cat(out_original_labels)

        return shadow_in_data, shadow_out_data, in_original_labels, out_original_labels

    def get_score(self):
        pass


class AttackerModel:
    def __init__(self, models):
        """train attack model for memnership inference
           reference https://arxiv.org/abs/1610.05820

        Args:
            models: models of attacker

        Attriutes:
            _init_models
            models
        """
        self._init_models = models
        self.models = {}

    def fit(self, shadow_result):
        """train an attacl model with the result of shadow models

        Args:
            shadow_result (dict): key is each class
                                  value is (shadow_data, shadow_label)
        """
        for model_idx, (label, (X, y)) in enumerate(shadow_result.items()):
            model = self._init_models[model_idx]
            X = np.array(X.cpu())
            y = np.array(y.cpu())
            model.fit(X, y)
            self.models[label] = model

    def predict(self, y_pred_prob, y_labels):
        """predict whether the given prediction came from training data or not

        Args:
            y_pred_prob (torch.Tensor): predicted probabilities on the data
            y_labels (torch.Tensor): true label of the data which y_pred_prob
                                     is predicted on

        Returns:
            in_out_pred (np.array) : result of attack
                                     each element should be one or zero
        """
        y_pred_prob = np.array(y_pred_prob.cpu())
        y_labels = np.array(y_labels.cpu())
        unique_labels = np.unique(y_labels)
        in_out_pred = np.zeros_like(y_labels)
        for label in unique_labels:
            idx = np.where(y_labels == label)
            in_out_pred[idx] = self.models[label].predict(y_pred_prob[idx])

        return in_out_pred

    def predict_proba(self, y_pred_prob, y_labels):
        """get probabilities of whether the given prediction came from
           training data or not

        Args:
            y_pred_prob (torch.Tensor): predicted probabilities on the data
            y_labels (torch.Tensor): true label of the data which y_pred_prob
                                     is predicted on

        Returns:
            in_out_pred (np.array) : result of attack
                                     each element expresses the possibility
        """
        y_pred_prob = np.array(y_pred_prob.cpu())
        y_labels = np.array(y_labels.cpu())
        unique_labels = np.unique(y_labels)
        in_out_pred = np.zeros_like(y_labels).astype(float)
        for label in unique_labels:
            idx = np.where(y_labels == label)
            in_out_pred[idx] = self.models[label].predict_proba(y_pred_prob[idx])[:, 1]

        return in_out_pred


class Membership_Inference(BaseAttacker):
    def __init__(
        self,
        target_model,
        shadow_models,
        attack_models,
        shadow_data_size,
        shadow_transform,
    ):
        """Implementation of membership inference
           reference https://arxiv.org/abs/1610.05820

        Args:
            target_model: the model of the victim
            shadow_models: shadow model for attack
            attack_models: attacker model for attack
            shadow_data_size: the size of datasets for
                              training the shadow models
            shadow_transform: the transformation function for shadow datasets

        Attributes:
            shadow_models
            attack_models
            shadow_data_size
            shadow_trnasform
            sm
            shadow_result
            am
        """
        super().__init__(target_model)
        self.shadow_models = shadow_models
        self.attack_models = attack_models
        self.shadow_data_size = shadow_data_size
        self.shadow_trasform = shadow_transform

        self.sm = None
        self.shadow_result = None
        self.am = None

    def train_shadow(self, X, y, num_itr):
        """train shadow models

        Args:
            X (np.array): training data for shadow models
            y (np.array): training label for shadow models
            num_itr (int): number of iteration for training
        """
        self.sm = ShadowModel(
            self.shadow_models,
            self.shadow_data_size,
            shadow_transform=self.shadow_trasform,
        )
        self.shadow_result = self.sm.fit_transform(X, y, num_itr=num_itr)

    def train_attacker(self):
        """Train attacker models"""
        self.am = AttackerModel(self.attack_models)
        self.am.fit(self.shadow_result)

    def attack(self, x, y, proba=False):
        """Attack victim model

        Args:
            x: target datasets which the attacker wants to classify
            y: target labels which the attacker wants to classify
            proba: the format of the output
        """
        prediction_of_taregt_model = self.target_model(x)
        if proba:
            return self.predit_proba(prediction_of_taregt_model, y)
        else:
            return self.predit(prediction_of_taregt_model, y)

    def predict(self, pred, label):
        """Predict whether the given prediction came from training data or not

        Args:
            pred (torch.Tensor): predicted probabilities on the data
            label (torch.Tensor): true label of the data which y_pred_prob
                                     is predicted on

        Returns:
            predicted binaru labels
        """
        return self.am.predict(pred, label)

    def predict_proba(self, pred, label):
        """get probabilities of whether the given prediction came from
           training data or not

        Args:
            pred (torch.Tensor): predicted probabilities on the data
            label (torch.Tensor): true label of the data which y_pred_prob
                                     is predicted on

        Returns:
            predicted probabilities
        """
        return self.am.predict_proba(pred, label)



import random

import torch

from ...manager import BaseManager


def attach_dba_to_client(
    cls, decomposed_trigger_rules, target_label, poison_ratio, scale
):
    """Wraps the given class in DistributedBackdoorAttackClientWrapper.

    Args:
        cls: Server class
        decomposed_trigger_rules ([function]): list of functions that define the decomposed trigger rules for each client
        target_label (int): a label that the attacker want to make the victim model predict when the inupt contains the trigger
        poison_ratio (float): a ratio of poisoned samples
        scale (_type_): scale for the uploaded gradients

    Returns:
        cls: a class wrapped in DistributedBackdoorAttackClientWrapper
    """

    class DistributedBackdoorAttackClientWrapper(cls):
        """Implementation of https://openreview.net/forum?id=rkgyS0VFvr"""

        def __init__(self, *args, **kwargs):
            super(DistributedBackdoorAttackClientWrapper, self).__init__(
                *args, **kwargs
            )

        def upload_gradients(self):
            """Uploads the local gradients"""
            gradients = []
            for param, prev_param in zip(self.model.parameters(), self.prev_parameters):
                gradients.append((prev_param - param) / self.lr * scale)
            return gradients

        def local_train(
            self, local_epoch, criterion, trainloader, optimizer, communication_id=0
        ):
            for i in range(local_epoch):
                running_loss = 0.0
                running_data_num = 0
                for _, data in enumerate(trainloader, 0):
                    inputs, labels = data
                    inputs = inputs.to(self.device)

                    if random.random() < poison_ratio:
                        inputs = decomposed_trigger_rules[self.user_id](inputs)
                        labels = torch.ones_like(labels) * target_label

                    labels = labels.to(self.device)

                    optimizer.zero_grad()
                    self.zero_grad()

                    outputs = self(inputs)
                    loss = criterion(outputs, labels)

                    loss.backward()
                    optimizer.step()

                    running_loss += loss.item()
                    running_data_num += inputs.shape[0]

                print(
                    f"communication {communication_id}, epoch {i}: client-{self.user_id+1}",
                    running_loss / running_data_num,
                )

    return DistributedBackdoorAttackClientWrapper


class DistributedBackdoorAttackClientManager(BaseManager):
    """Manager class for DistributedBackdoorAttack proposed in https://openreview.net/forum?id=rkgyS0VFvr."""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        """Wraps the given class in DistributedBackdoorAttackClientWrapper.

        Returns:
            cls: a class wrapped in DistributedBackdoorAttackClientWrapper
        """
        return attach_dba_to_client(cls, *self.args, **self.kwargs)



import numpy as np
import torch


def transform_list_to_tensor(model_params_list):
    for k in model_params_list.keys():
        model_params_list[k] = torch.from_numpy(
            np.asarray(model_params_list[k])
        ).float()
    return model_params_list



"""Submodule for attack algorithms against machine learning.
"""
from .base_attack import BaseAttacker  # noqa: F401
from .evasion import Evasion_attack_sklearn  # noqa: F401
from .inversion import (  # noqa: F401
    MI_FACE,
    GANAttackClientManager,
    Generator_Attack,
    GradientInversion_Attack,
    GradientInversionAttackServerManager,
    attach_ganattack_to_client,
)
from .labelleakage import (  # noqa: F401
    NormAttackSplitNNManager,
    attach_normattack_to_splitnn,
)
from .membership import AttackerModel, Membership_Inference, ShadowModel  # noqa: F401
from .poison import Poison_attack_sklearn  # noqa: F401



from abc import ABCMeta, abstractmethod


class BaseAttacker(metaclass=ABCMeta):
    def __init__(self, target_model):
        """attacker against ml model

        Args:
            target_model: target ML model
        """
        self.target_model = target_model

    @abstractmethod
    def attack(self):
        pass



import copy

import torch
import torch.nn as nn

from ...manager import BaseManager
from ..base_attack import BaseAttacker
from .utils.distance import cossim, l2
from .utils.regularization import (
    bn_regularizer,
    group_consistency,
    label_matching,
    total_variance,
)


class GradientInversion_Attack(BaseAttacker):
    """General Gradient Inversion Attacker

    model inversion attack based on gradients can be written as follows:
            x^* = argmin_x' L_grad(x': W, delta_W) + R_aux(x')
    , where X' is the reconstructed image.
    The attacker tries to find images whose gradients w.r.t the given model parameter W
    is similar to the gradients delta_W of the secret images.

    Attributes:
        target_model: a target torch module instance.
        x_shape: the input shape of target_model.
        y_shape: the output shape of target_model.
        optimize_label: If true, only optimize images (the label will be automatically estimated).
        pos_of_final_fc_layer: position of gradients corresponding to the final FC layer
                               within the gradients received from the client.
        num_iteration: number of iterations of optimization.
        optimizer_class: a class of torch optimizer for the attack.
        lossfunc: a function that takes the predictions of the target model and true labels
                  and returns the loss between them.
        distancefunc: a function which takes the gradients of reconstructed images and the client-side gradients
                      and returns the distance between them.
        tv_reg_coef: the coefficient of total-variance regularization.
        lm_reg_coef: the coefficient of label-matching regularization.
        l2_reg_coef: the coefficient of L2 regularization.
        bn_reg_coef: the coefficient of BN regularization.
        gc_reg_coef: the coefficient of group-consistency regularization.
        bn_reg_layers: a list of batch normalization layers of the target model.
        bn_reg_layer_inputs: a lit of extracted inputs of the specified bn layers
        custom_reg_func: a custom regularization function.
        custom_reg_coef: the coefficient of the custom regularization function
        device: device type.
        log_interval: the interval of logging.
        save_loss: If true, save the loss during the attack.
        seed: random state.
        group_num: the size of group,
        group_seed: a list of random states for each worker of the group
        early_stopping: early stopping
    """

    def __init__(
        self,
        target_model,
        x_shape,
        y_shape=None,
        optimize_label=True,
        gradient_ignore_pos=[],
        pos_of_final_fc_layer=-2,
        num_iteration=100,
        optimizer_class=torch.optim.LBFGS,
        optimizername=None,
        lossfunc=nn.CrossEntropyLoss(),
        distancefunc=l2,
        distancename=None,
        tv_reg_coef=0.0,
        lm_reg_coef=0.0,
        l2_reg_coef=0.0,
        bn_reg_coef=0.0,
        gc_reg_coef=0.0,
        bn_reg_layers=[],
        custom_reg_func=None,
        custom_reg_coef=0.0,
        custom_generate_fake_grad_fn=None,
        device="cpu",
        log_interval=10,
        save_loss=True,
        seed=0,
        group_num=5,
        group_seed=None,
        early_stopping=50,
        clamp_range=None,
        **kwargs,
    ):
        """Inits GradientInversion_Attack class.

        Args:
            target_model: a target torch module instance.
            x_shape: the input shape of target_model.
            y_shape: the output shape of target_model.
            optimize_label: If true, only optimize images (the label will be automatically estimated).
            gradient_ignore_pos: a list of positions whihc will be ignored during the culculation of
                                 the distance between gradients
            pos_of_final_fc_layer: position of gradients corresponding to the final FC layer
                                   within the gradients received from the client.
            num_iteration: number of iterations of optimization.
            optimizer_class: a class of torch optimizer for the attack.
            optimizername: a name of optimizer class (priority over optimizer_class).
            lossfunc: a function that takes the predictions of the target model and true labels
                    and returns the loss between them.
            distancefunc: a function which takes the gradients of reconstructed images and the client-side gradients
                        and returns the distance between them.
            distancename: a name of distancefunc (priority over distancefunc).
            tv_reg_coef: the coefficient of total-variance regularization.
            lm_reg_coef: the coefficient of label-matching regularization.
            l2_reg_coef: the coefficient of L2 regularization.
            bn_reg_coef: the coefficient of BN regularization.
            gc_reg_coef: the coefficient of group-consistency regularization.
            bn_reg_layers: a list of batch normalization layers of the target model.
            custom_reg_func: a custom regularization function.
            custom_reg_coef: the coefficient of the custom regularization function
            device: device type.
            log_interval: the interval of logging.
            save_loss: If true, save the loss during the attack.
            seed: random state.
            group_num: the size of group,
            group_seed: a list of random states for each worker of the group
            early_stopping: early stopping
            **kwargs: kwargs for the optimizer
        """
        super().__init__(target_model)
        self.x_shape = x_shape
        self.y_shape = (
            list(target_model.parameters())[-1].shape[0] if y_shape is None else y_shape
        )

        self.optimize_label = optimize_label
        self.gradient_ignore_pos = gradient_ignore_pos
        self.pos_of_final_fc_layer = pos_of_final_fc_layer

        self.num_iteration = num_iteration
        self.lossfunc = lossfunc
        self.distancefunc = distancefunc
        self._setup_distancefunc(distancename)
        self.optimizer_class = optimizer_class
        self._setup_optimizer_class(optimizername)

        self.tv_reg_coef = tv_reg_coef
        self.lm_reg_coef = lm_reg_coef
        self.l2_reg_coef = l2_reg_coef
        self.bn_reg_coef = bn_reg_coef
        self.gc_reg_coef = gc_reg_coef

        self.bn_reg_layers = bn_reg_layers
        self.bn_reg_layer_inputs = {}
        for i, bn_layer in enumerate(self.bn_reg_layers):
            bn_layer.register_forward_hook(self._get_hook_for_input(i))

        self.custom_reg_func = custom_reg_func
        self.custom_reg_coef = custom_reg_coef

        self.custom_generate_fake_grad_fn = custom_generate_fake_grad_fn

        self.device = device
        self.log_interval = log_interval
        self.save_loss = save_loss
        self.seed = seed

        self.group_num = group_num
        self.group_seed = list(range(group_num)) if group_seed is None else group_seed

        self.early_stopping = early_stopping
        self.clamp_range = clamp_range

        self.kwargs = kwargs

        torch.manual_seed(seed)

    def _setup_distancefunc(self, distancename):
        """Assigns a function to self.distancefunc according to distancename

        Args:
            distancename: name of the function to culculat the distance between the gradients.
                          currently support 'l2' or 'cossim'.

        Raises:
            ValueError: if distancename is not supported.
        """
        if distancename is None:
            return
        elif distancename == "l2":
            self.distancefunc = l2
        elif distancename == "cossim":
            self.distancefunc = cossim
        else:
            raise ValueError(f"{distancename} is not defined")

    def _setup_optimizer_class(self, optimizername):
        """Assigns a class to self.optimizer_class according to optimiername

        Args:
            optimizername: name of optimizer, currently support `LBFGS`, `SGD`, and `Adam`

        Raises:
            ValueError: if optimizername is not supported.
        """
        if optimizername is None:
            return
        elif optimizername == "LBFGS":
            self.optimizer_class = torch.optim.LBFGS
        elif optimizername == "SGD":
            self.optimizer_class = torch.optim.SGD
        elif optimizername == "Adam":
            self.optimizer_class = torch.optim.Adam
        else:
            raise ValueError(f"{optimizername} is not defined")

    def _get_hook_for_input(self, name):
        """Returns a hook function to extract the input of the specified layer of the target model

        Args:
            name: the key of self.bn_reg_layer_inputs for the target layer

        Returns:
            hook: a hook function
        """

        def hook(model, inp, output):
            self.bn_reg_layer_inputs[name] = inp[0]

        return hook

    def _initialize_x(self, batch_size):
        """Inits the fake images

        Args:
            batch_size: the batch size

        Returns:
            randomly generated torch.Tensor whose shape is (batch_size, ) + (self.x_shape)
        """
        fake_x = torch.randn(
            (batch_size,) + (self.x_shape), requires_grad=True, device=self.device
        )
        return fake_x

    def _initialize_label(self, batch_size):
        """Inits the fake labels

        Args:
            batch_size: the batch size

        Returns:
            randomly initialized or estimated labels
        """
        fake_label = torch.randn(
            (batch_size, self.y_shape), requires_grad=True, device=self.device
        )
        fake_label = fake_label.to(self.device)
        return fake_label

    def _estimate_label(self, received_gradients, batch_size):
        """Estimates the secret labels from the received gradients

        this function is based on the following papers:
        batch_size == 1: https://arxiv.org/abs/2001.02610
        batch_size > 1: https://arxiv.org/abs/2104.07586

        Args:
            received_gradients: gradients received from the client
            batch_size: batch size used to culculate the received_gradients

        Returns:
            estimated labels
        """
        if batch_size == 1:
            fake_label = torch.argmin(
                torch.sum(received_gradients[self.pos_of_final_fc_layer], dim=1)
            )
        else:
            fake_label = torch.argsort(
                torch.min(received_gradients[self.pos_of_final_fc_layer], dim=-1)[0]
            )[:batch_size]
        fake_label = fake_label.reshape(batch_size)
        fake_label = fake_label.to(self.device)
        return fake_label

    def _culc_regularization_term(
        self, fake_x, fake_pred, fake_label, group_fake_x, received_gradients
    ):
        """Culculates the regularization term

        Args:
            fake_x: reconstructed images
            fake_pred: the predicted value of reconstructed images
            faka_label: the labels of fake_x
            group_fake_x: a list of fake_x of each worker
            received_gradients: gradients received from the client

        Returns:
            culculated regularization term
        """
        reg_term = 0
        if self.tv_reg_coef != 0:
            reg_term += self.tv_reg_coef * total_variance(fake_x)
        if self.lm_reg_coef != 0:
            reg_term += self.lm_reg_coef * label_matching(fake_pred, fake_label)
        if self.l2_reg_coef != 0:
            reg_term += self.l2_reg_coef * torch.norm(fake_x, p=2)
        if self.bn_reg_coef != 0:
            reg_term += self.bn_reg_coef * bn_regularizer(
                self.bn_reg_layer_inputs, self.bn_reg_layers
            )
        if group_fake_x is not None and self.gc_reg_coef != 0:
            reg_term += self.gc_reg_coef * group_consistency(fake_x, group_fake_x)
        if self.custom_reg_func is not None and self.custom_reg_coef != 0:
            context = {
                "attacker": self,
                "fake_x": fake_x,
                "fake_label": fake_label,
                "received_gradients": received_gradients,
                "group_fake_x": group_fake_x,
            }
            reg_term += self.custom_reg_coef * self.custom_reg_func(context)

        return reg_term

    def _generate_fake_gradients(self, fake_x, fake_label):
        fake_pred = self.target_model(fake_x)
        if self.optimize_label:
            loss = self.lossfunc(fake_pred, fake_label.softmax(dim=-1))
        else:
            loss = self.lossfunc(fake_pred, fake_label)
        fake_gradients = torch.autograd.grad(
            loss,
            self.target_model.parameters(),
            create_graph=True,
            allow_unused=True,
        )
        return fake_pred, fake_gradients

    def _setup_closure(
        self, optimizer, fake_x, fake_label, received_gradients, group_fake_x=None
    ):
        """Returns a closure function for the optimizer

        Args:
            optimizer (torch.optim.Optimizer): an instance of the optimizer
            fake_x (torch.Tensor): reconstructed images
            fake_label (torch.Tensor): reconstructed or estimated labels
            received_gradients (list): a list of gradients received from the client
            group_fake_x (list, optional): a list of fake_x. Defaults to None.
        """

        def closure():
            if self.custom_generate_fake_grad_fn is None:
                fake_pred, fake_gradients = self._generate_fake_gradients(
                    fake_x, fake_label
                )
            else:
                fake_pred, fake_gradients = self.custom_generate_fake_grad_fn(
                    self, fake_x, fake_label
                )
            optimizer.zero_grad()
            distance = self.distancefunc(
                fake_gradients, received_gradients, self.gradient_ignore_pos
            )
            distance += self._culc_regularization_term(
                fake_x,
                fake_pred,
                fake_label,
                group_fake_x,
                received_gradients,
            )

            distance.backward(retain_graph=True)
            return distance

        return closure

    def _setup_attack(self, received_gradients, batch_size, init_x=None, labels=None):
        """Initializes the image and label, and set the optimizer

        Args:
            received_gradients: a list of gradients received from the client
            batch_size: the batch size

        Returns:
            initial images, labels, and the optimizer instance
        """
        fake_x = self._initialize_x(batch_size) if init_x is None else init_x

        if labels is None:
            fake_label = (
                self._initialize_label(batch_size)
                if self.optimize_label
                else self._estimate_label(received_gradients, batch_size)
            )
        else:
            fake_label = labels

        optimizer = (
            self.optimizer_class([fake_x, fake_label], **self.kwargs)
            if self.optimize_label
            else self.optimizer_class(
                [
                    fake_x,
                ],
                **self.kwargs,
            )
        )

        return fake_x, fake_label, optimizer

    def reset_seed(self, seed):
        """Resets the random seed

        Args:
            seed (int): the random seed
        """
        self.seed = seed
        torch.manual_seed(seed)

    def attack(
        self,
        received_gradients,
        batch_size=1,
        init_x=None,
        labels=None,
        return_best=True,
    ):
        """Reconstructs the images from the gradients received from the client

        Args:
            received_gradients: the list of gradients received from the client.
            batch_size: batch size.

        Returns:
            a tuple of the best reconstructed images and corresponding labels

        Raises:
            ValueError: If the culculated distance become Nan
        """
        fake_x, fake_label, optimizer = self._setup_attack(
            received_gradients, batch_size, init_x=init_x, labels=labels
        )

        num_of_not_improve_round = 0
        best_distance = float("inf")
        self.log_loss = []
        for i in range(1, self.num_iteration + 1):
            closure = self._setup_closure(
                optimizer, fake_x, fake_label, received_gradients
            )
            distance = optimizer.step(closure)

            if self.clamp_range is not None:
                with torch.no_grad():
                    fake_x[:] = fake_x.clamp(self.clamp_range[0], self.clamp_range[1])

            if torch.sum(torch.isnan(distance)).item():
                raise ValueError("stop because the culculated distance is Nan")

            if self.save_loss:
                self.log_loss.append(distance)

            if best_distance > distance:
                if return_best:
                    best_fake_x = copy.deepcopy(fake_x)
                    best_fake_label = copy.deepcopy(fake_label)
                best_distance = distance
                best_iteration = i
                num_of_not_improve_round = 0
            else:
                num_of_not_improve_round += 1

            if self.log_interval != 0 and i % self.log_interval == 0:
                print(
                    f"iter={i}: {distance}, (best_iter={best_iteration}: {best_distance})"
                )

            if num_of_not_improve_round > self.early_stopping:
                print(
                    f"iter={i}: loss did not improve in the last {self.early_stopping} rounds."
                )
                break

        if return_best:
            return best_fake_x, best_fake_label
        else:
            return fake_x, fake_label

    def group_attack(self, received_gradients, batch_size=1):
        """Multiple simultaneous attacks with different random states

        Args:
            received_gradients: the list of gradients received from the client.
            batch_size: batch size.

        Returns:
            a tuple of the best reconstructed images and corresponding labels

        Raises:
            ValueError: If the culculated distance become Nan
        """
        group_fake_x = []
        group_fake_label = []
        group_optimizer = []

        for _ in range(self.group_num):
            fake_x, fake_label, optimizer = self._setup_attack(
                received_gradients, batch_size
            )
            group_fake_x.append(fake_x)
            group_fake_label.append(fake_label)
            group_optimizer.append(optimizer)

        best_distance = [float("inf") for _ in range(self.group_num)]
        best_fake_x = copy.deepcopy(group_fake_x)
        best_fake_label = copy.deepcopy(group_fake_label)
        best_iteration = [0 for _ in range(self.group_num)]

        self.log_loss = [[] for _ in range(self.group_num)]
        for i in range(1, self.num_iteration + 1):
            for worker_id in range(self.group_num):
                self.reset_seed(self.group_seed[worker_id])
                closure = self._setup_closure(
                    group_optimizer[worker_id],
                    group_fake_x[worker_id],
                    group_fake_label[worker_id],
                    received_gradients,
                )
                distance = group_optimizer[worker_id].step(closure)

                if self.save_loss:
                    self.log_loss[worker_id].append(distance)

                if best_distance[worker_id] > distance:
                    best_fake_x[worker_id] = copy.deepcopy(group_fake_x[worker_id])
                    best_fake_label[worker_id] = copy.deepcopy(
                        group_fake_label[worker_id]
                    )
                    best_distance[worker_id] = distance
                    best_iteration[worker_id] = i

                if self.log_interval != 0 and i % self.log_interval == 0:
                    print(
                        f"worker_id={worker_id}: iter={i}: {distance}, (best_iter={best_iteration[worker_id]}: {best_distance[worker_id]})"
                    )

        return best_fake_x, best_fake_label


def _default_gradinent_inversion_attack_on_receive(self):
    tmp_result = []
    for s in range(self.num_trial_per_communication):
        self.reset_seed(s)
        try:
            tmp_result.append(self.attack())
        except:
            continue
    self.attack_results.append(tmp_result)


def attach_gradient_inversion_attack_to_server(
    cls,
    x_shape,
    y_shape=None,
    attack_function_on_receive=_default_gradinent_inversion_attack_on_receive,
    num_trial_per_communication=1,
    optimize_label=True,
    gradient_ignore_pos=[],
    pos_of_final_fc_layer=-2,
    num_iteration=100,
    optimizer_class=torch.optim.LBFGS,
    optimizername=None,
    lossfunc=nn.CrossEntropyLoss(),
    distancefunc=l2,
    distancename=None,
    tv_reg_coef=0.0,
    lm_reg_coef=0.0,
    l2_reg_coef=0.0,
    bn_reg_coef=0.0,
    gc_reg_coef=0.0,
    bn_reg_layers=[],
    custom_reg_func=None,
    custom_reg_coef=0.0,
    custom_generate_fake_grad_fn=None,
    device="cpu",
    log_interval=10,
    save_loss=True,
    seed=0,
    group_num=5,
    group_seed=None,
    early_stopping=50,
    clamp_range=None,
    target_client_id=0,
    gradinvattack_kwargs={},
):
    """Wraps the given class in GradientInversionServerWrapper.

    Args:
        x_shape: input shape of target_model.
        y_shape: output shape of target_model.
        attack_function_on_receive (function, optional): a function to execute attack called after receving the local gradients. Defaults to _default_gradinent_inversion_attack_on_receive.
        num_trial_per_communication (int, optional): number of attack trials executed per communication. Defaults to 1.
        optimize_label: If true, only optimize images (the label will be automatically estimated).
        gradient_ignore_pos: a list of positions whihc will be ignored during the culculation of
                                the distance between gradients
        pos_of_final_fc_layer: position of gradients corresponding to the final FC layer
                                within the gradients received from the client.
        num_iteration: number of iterations of optimization.
        optimizer_class: a class of torch optimizer for the attack.
        optimizername: a name of optimizer class (priority over optimizer_class).
        lossfunc: a function that takes the predictions of the target model and true labels
                and returns the loss between them.
        distancefunc: a function which takes the gradients of reconstructed images and the client-side gradients
                    and returns the distance between them.
        distancename: a name of distancefunc (priority over distancefunc).
        tv_reg_coef: the coefficient of total-variance regularization.
        lm_reg_coef: the coefficient of label-matching regularization.
        l2_reg_coef: the coefficient of L2 regularization.
        bn_reg_coef: the coefficient of BN regularization.
        gc_reg_coef: the coefficient of group-consistency regularization.
        bn_reg_layers: a list of batch normalization layers of the target model.
        custom_reg_func: a custom regularization function.
        custom_reg_coef: the coefficient of the custom regularization function
        device: device type.
        log_interval: the interval of logging.
        save_loss: If true, save the loss during the attack.
        seed: random state.
        group_num: the size of group,
        group_seed: a list of random states for each worker of the group
        early_stopping: early stopping
        clamp_range ((float, float), optional): clamps the fake input within the specified range. Defaults to None.
        target_client_id (int, optional): id f the target client. Defaults to 0.
        gradinvattack_kwargs (dict, optional): optional arguments for the optimizer used in inversion attack. Defaults to {}.

    Returns:
        cls: GradientInversionServerWrapper
    """

    class GradientInversionServerWrapper(cls):
        def __init__(self, *args, **kwargs):
            super(GradientInversionServerWrapper, self).__init__(*args, **kwargs)
            self.target_client_id = target_client_id
            self.num_trial_per_communication = num_trial_per_communication
            self.attacker = GradientInversion_Attack(
                self.server_model,
                x_shape,
                y_shape=y_shape,
                optimize_label=optimize_label,
                gradient_ignore_pos=gradient_ignore_pos,
                pos_of_final_fc_layer=pos_of_final_fc_layer,
                num_iteration=num_iteration,
                optimizer_class=optimizer_class,
                optimizername=optimizername,
                lossfunc=lossfunc,
                distancefunc=distancefunc,
                distancename=distancename,
                tv_reg_coef=tv_reg_coef,
                lm_reg_coef=lm_reg_coef,
                l2_reg_coef=l2_reg_coef,
                bn_reg_coef=bn_reg_coef,
                gc_reg_coef=gc_reg_coef,
                bn_reg_layers=bn_reg_layers,
                custom_reg_func=custom_reg_func,
                custom_reg_coef=custom_reg_coef,
                custom_generate_fake_grad_fn=custom_generate_fake_grad_fn,
                device=device,
                log_interval=log_interval,
                save_loss=save_loss,
                seed=seed,
                group_num=group_num,
                group_seed=group_seed,
                early_stopping=early_stopping,
                clamp_range=clamp_range,
                **gradinvattack_kwargs,
            )

            self.attack_results = []

        def change_target_client_id(self, target_client_id):
            self.target_client_id = target_client_id
            self.attacker.target_model = self.clients[self.target_client_id]

        def receive(self, *args, **kwargs):
            super(GradientInversionServerWrapper, self).receive(*args, **kwargs)

            attack_function_on_receive(self)

        def attack(self, **kwargs):
            received_gradient = self.uploaded_gradients[self.target_client_id]
            return self.attacker.attack(received_gradient, **kwargs)

        def group_attack(self, **kwargs):
            received_gradient = self.uploaded_gradients[self.target_client_id]
            return self.attacker.group_attack(received_gradient, **kwargs)

        def reset_seed(self, seed):
            self.attacker.reset_seed(seed)

    return GradientInversionServerWrapper


class GradientInversionAttackServerManager(BaseManager):
    """Manager class for Gradient-based model inversion attack"""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        """Wraps the given class in GradientInversionServerWrapper.

        Returns:
            cls: GradientInversionServerWrapper
        """
        return attach_gradient_inversion_attack_to_server(
            cls, *self.args, **self.kwargs
        )



import copy

from ..fedavg import FedAVGAPI, MPIFedAVGAPI


class FedProxAPI(FedAVGAPI):
    """Implementation of FedProx (https://arxiv.org/abs/1812.06127)"""

    def __init__(self, *args, mu=0.01, **kwargs):
        super().__init__(*args, **kwargs)
        self.mu = mu

    def local_train(self, i):
        for client_idx in range(self.client_num):
            self.clients[client_idx].local_train(
                self.server.parameters(),
                self.local_epoch,
                self.criterion,
                self.local_dataloaders[client_idx],
                self.local_optimizers[client_idx],
                communication_id=i,
            )

    def run(self):
        for i in range(self.num_communication):
            self.local_train(i)

            self.server.receive(use_gradients=self.use_gradients)
            if self.use_gradients:
                self.server.updata_from_gradients(weight=self.clients_weight)
            else:
                self.server.update_from_parameters(weight=self.clients_weight)

            self.custom_action(self)


class MPIFedProxAPI(MPIFedAVGAPI):
    def __init__(self, *args, mu=0.01, **kwargs):
        super().__init__(*args, **kwargs)
        self.mu = mu

    def local_train(self, com_cnt):
        self.party.prev_parameters = []
        for param in self.party.parameters():
            self.party.prev_parameters.append(copy.deepcopy(param))

        self.party.local_train(
            self.party.prev_parameters,
            self.local_epoch,
            self.criterion,
            self.local_dataloader,
            self.local_optimizer,
            communication_id=com_cnt,
        )



import torch
from torch import nn

from ...manager import BaseManager
from ...utils.utils import torch_round_x_decimal
from ..core import BaseClient
from ..core.utils import GLOBAL_LOGIT_TAG, LOCAL_LOGIT_TAG


class FedMDClient(BaseClient):
    def __init__(
        self,
        model,
        public_dataloader,
        output_dim=1,
        batch_size=8,
        user_id=0,
        base_loss_func=nn.CrossEntropyLoss(),
        consensus_loss_func=nn.L1Loss(),
        round_decimal=None,
        device="cpu",
    ):
        super(FedMDClient, self).__init__(model, user_id=user_id)
        self.public_dataloader = public_dataloader
        self.batch_size = batch_size
        self.base_loss_func = base_loss_func
        self.consensus_loss_func = consensus_loss_func
        self.round_decimal = round_decimal
        self.device = device

        self.predicted_values_of_server = None

        len_public_dataloader = len(self.public_dataloader.dataset)
        self.logit2server = torch.ones((len_public_dataloader, output_dim)).to(
            self.device
        ) * float("inf")

    def upload(self):
        for data in self.public_dataloader:
            idx = data[0]
            x = data[1]
            x = x.to(self.device)
            self.logit2server[idx, :] = self(x).detach()

        if self.round_decimal is None:
            return self.logit2server
        else:
            return torch_round_x_decimal(self.logit2server, self.round_decimal)

    def download(self, predicted_values_of_server):
        self.predicted_values_of_server = predicted_values_of_server

    def local_train(self, local_epoch, criterion, trainloader, optimizer):

        running_loss = 0.0
        for _ in range(local_epoch):
            for data in trainloader:
                _, x, y = data
                x = x.to(self.device)
                y = y.to(self.device).to(torch.int64)

                optimizer.zero_grad()
                loss = criterion(self(x), y)
                loss.backward()
                optimizer.step()

                running_loss += loss.item()

        return running_loss

    def approach_consensus(self, consensus_optimizer):
        running_loss = 0

        for data in self.public_dataloader:
            idx = data[0]
            x = data[1].to(self.device)
            y_consensus = self.predicted_values_of_server[idx, :].to(self.device)
            consensus_optimizer.zero_grad()
            y_pred = self(x)
            loss = self.consensus_loss_func(y_pred, y_consensus)
            loss.backward()
            consensus_optimizer.step()
            running_loss += loss.item()

        return running_loss


def attach_mpi_to_fedmdclient(cls):
    class MPIFedMDClientWrapper(cls):
        def __init__(self, comm, *args, **kwargs):
            super(MPIFedMDClientWrapper, self).__init__(*args, **kwargs)
            self.comm = comm

        def action(self):
            self.mpi_upload()
            self.model.zero_grad()
            self.mpi_download()

        def mpi_upload(self):
            self.mpi_upload_logits()

        def mpi_upload_logits(self, destination_id=0):
            self.comm.send(self.upload(), dest=destination_id, tag=LOCAL_LOGIT_TAG)

        def mpi_download(self):
            self.download(self.comm.recv(tag=GLOBAL_LOGIT_TAG))

        def mpi_initialize(self):
            self.mpi_download()

    return MPIFedMDClientWrapper


class MPIFedMDClientManager(BaseManager):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        return attach_mpi_to_fedmdclient(cls, *self.args, **self.kwargs)



import torch

from ...manager import BaseManager


def attach_mapf_to_client(cls, lam, base_model_parameters=None):
    class MAPFClientWrapper(cls):
        """Implementation of MAPF proposed in https://arxiv.org/pdf/2203.08669.pdf"""

        def __init__(self, *args, **kwargs):
            super(MAPFClientWrapper, self).__init__(*args, **kwargs)

            if base_model_parameters is None:
                self.base_model_parameters = [
                    torch.randn_like(p) for p in self.model.parameters()
                ]
            else:
                self.base_model_parameters = base_model_parameters

        def upload_gradients(self):
            """Upload the local gradients"""
            gradients = []
            for param, base_param in zip(
                self.model.parameters(), self.base_model_parameters
            ):
                gradients.append((base_param - param) * lam)
            return gradients

    return MAPFClientWrapper


class MAPFClientWrapper(BaseManager):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        return attach_mapf_to_client(cls, *self.args, **self.kwargs)



import torch
from torch import nn

from ...utils.utils import torch_round_x_decimal
from ..core import BaseClient


class FedGEMSClient(BaseClient):
    def __init__(
        self,
        model,
        user_id=0,
        lr=0.1,
        base_loss_func=nn.CrossEntropyLoss(),
        kldiv_loss_func=nn.KLDivLoss(),
        epsilon=0.75,
        round_decimal=None,
    ):
        super(FedGEMSClient, self).__init__(model, user_id=user_id)
        self.lr = lr
        self.predicted_values_of_server = None
        self.base_loss_func = base_loss_func
        self.kldiv_loss_func = kldiv_loss_func
        self.epsilon = epsilon
        self.round_decimal = round_decimal

    def upload(self, x):
        result = self(x)
        if self.round_decimal is None:
            return result
        else:
            return torch_round_x_decimal(result, self.round_decimal)

    def download(self, predicted_values_of_server):
        self.predicted_values_of_server = predicted_values_of_server

    def local_train(self, local_epoch, criterion, trainloader, optimizer):

        running_loss = 0.0
        for _ in range(local_epoch):
            for data in trainloader:
                _, x, y = data
                x = x.to(self.device)
                y = y.to(self.device).to(torch.int64)

                optimizer.zero_grad()
                loss = criterion(self(x), y)
                loss.backward()
                optimizer.step()

                running_loss += loss.item()

        return running_loss

    def culc_loss_on_public_dataset(self, idx, y_pred, y):
        y_pred_server = self.predicted_values_of_server[idx]
        base_loss = self.epsilon * self.base_loss_func(y_pred, y.to(torch.int64))
        kl_loss = (1 - self.epsilon) * self.kldiv_loss_func(
            y_pred_server.softmax(dim=-1).log(), y_pred.softmax(dim=-1)
        )
        return base_loss + kl_loss



"""Implementation of DS-FL,  `Itahara, Sohei, et al. "Distillation-based semi-supervised federated learning for
communication-efficient collaborative training with non-iid private data.
" arXiv preprint arXiv:2008.06180 (2020).`"""
from .api import DSFLAPI  # noqa : F401
from .client import DSFLClient  # noqa : F401
from .server import DSFLServer  # noqa : F401



import copy

import torch

from ...manager import BaseManager


def attach_ganattack_to_client(
    cls,
    target_label,
    generator,
    generator_optimizer,
    generator_criterion,
    nz=100,
    device="cpu",
    gan_batch_size=1,
    gan_epoch=1,
    gan_log_interval=0,
    ignore_first_download=False,
):
    """Wraps the given class in GANAttackClientWrapper.

    Args:
        target_label(int): index of target class
        generator (torch.nn.Module): Generator
        generator_optimizer (torch.optim.Optimizer): optimizer for the generator
        generator_criterion (function): loss function for the generator
        nz (int): dimension of latent space of the generator. Defaults to 100.
        device (str, optional): _description_. Defaults to "cpu".
        gan_batch_size (int, optional): batch size for training GAN. Defaults to 1.
        gan_epoch (int, optional): epoch for training GAN. Defaults to 1.
        gan_log_interval (int, optional): log interval. Defaults to 0.
        ignore_first_download (bool, optional): Defaults to False.

    Returns:
        cls: a class wrapped in GANAttackClientWrapper
    """

    class GANAttackClientWrapper(cls):
        """Implementation of GAN based model inversion attack (https://arxiv.org/abs/1702.07464)"""

        def __init__(self, *args, **kwargs):
            super(GANAttackClientWrapper, self).__init__(*args, **kwargs)

            self.target_label = target_label
            self.generator = generator
            self.generator_optimizer = generator_optimizer
            self.generator_criterion = generator_criterion
            self.nz = nz
            self.device = device

            self.discriminator = copy.deepcopy(self.model)
            self.discriminator.to(self.device)

            self.noise = torch.randn(1, self.nz, 1, 1, device=self.device)
            self.is_params_initialized = False

        def update_generator(self, batch_size=10, epoch=1, log_interval=5):
            """Updates the Generator

            Args:
                batch_size (int): batch size
                epoch (int): epoch
                log_interval (int): interval of logging
            """

            for i in range(1, epoch + 1):
                running_error = 0
                self.generator.zero_grad()

                noise = torch.randn(batch_size, self.nz, 1, 1, device=self.device)
                fake = self.generator(noise)
                output = self.discriminator(fake)

                label = torch.full(
                    (batch_size,),
                    self.target_label,
                    dtype=torch.int64,
                    device=self.device,
                )
                loss_generator = self.generator_criterion(output, label)
                loss_generator.backward()

                self.generator_optimizer.step()

                running_error += loss_generator.item()

                if log_interval != 0 and i % log_interval == 0:
                    print(
                        f"updating generator - epoch {i}: generator loss is {running_error/batch_size}"
                    )

        def update_discriminator(self):
            """Updates the discriminator"""
            self.discriminator.load_state_dict(self.model.state_dict())

        def download(self, model_parameters):
            """Downloads the new model"""
            super().download(model_parameters)
            if ignore_first_download and not self.is_params_initialized:
                self.is_params_initialized = True
                return
            self.update_discriminator()
            self.update_generator(
                batch_size=gan_batch_size,
                epoch=gan_epoch,
                log_interval=gan_log_interval,
            )

        def attack(self, n):
            """Generates fake images

            Args:
                n (int): the number of fake images created by the Generator

            Returns:
                fake: generated fake images
            """
            noise = torch.randn(n, self.nz, 1, 1, device=self.device)
            with torch.no_grad():
                fake = self.generator(noise)
            return fake

    return GANAttackClientWrapper


class GANAttackClientManager(BaseManager):
    """Manager class for GAN based model inversion attack (https://arxiv.org/abs/1702.07464)"""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        """Wraps the given class in GANAttackClientWrapper.

        Returns:
            cls: a class wrapped in GANAttackClientWrapper
        """
        return attach_ganattack_to_client(cls, *self.args, **self.kwargs)



import numpy as np
import torch

from ...manager import BaseManager
from ..core import BaseServer
from ..core.utils import GRADIENTS_TAG, PARAMETERS_TAG
from ..optimizer import AdamFLOptimizer, SGDFLOptimizer


class FedAVGServer(BaseServer):
    """Server of FedAVG for single process simulation

    Args:
        clients ([FedAvgClient] | [int]): a list of FedAVG clients or their ids.
        global_model (torch.nn.Module): global model.
        server_id (int, optional): id of this server. Defaults to 0.
        lr (float, optional): learning rate. Defaults to 0.1.
        optimizer_type (str, optional): optimizer for the update of global model . Defaults to "sgd".
        server_side_update (bool, optional): If True, update the global model at the server-side. Defaults to True.
        optimizer_kwargs (dict, optional): kwargs for the global optimizer. Defaults to {}.
    """

    def __init__(
        self,
        clients,
        global_model,
        server_id=0,
        lr=0.1,
        optimizer_type="sgd",
        server_side_update=True,
        optimizer_kwargs={},
        device="cpu",
    ):
        super(FedAVGServer, self).__init__(clients, global_model, server_id=server_id)
        self.lr = lr
        self._setup_optimizer(optimizer_type, **optimizer_kwargs)
        self.server_side_update = server_side_update
        self.device = device
        self.uploaded_gradients = []

        self.force_send_model_state_dict = True

    def _setup_optimizer(self, optimizer_type, **kwargs):
        if optimizer_type == "sgd":
            self.optimizer = SGDFLOptimizer(
                self.server_model.parameters(), lr=self.lr, **kwargs
            )
        elif optimizer_type == "adam":
            self.optimizer = AdamFLOptimizer(
                self.server_model.parameters(), lr=self.lr, **kwargs
            )
        elif optimizer_type == "none":
            self.optimizer = None
        else:
            raise NotImplementedError(
                f"{optimizer_type} is not supported. You can specify `sgd`, `adam`, or `none`."
            )

    def action(self, use_gradients=True):
        self.receive(use_gradients)
        self.update(use_gradients)
        self.distribute()

    def receive(self, use_gradients=True):
        """Receive the local models

        Args:
            use_gradients (bool, optional): If True, receive the local gradients. Otherwise, receive the local parameters. Defaults to True.
        """
        if use_gradients:
            self.receive_local_gradients()
        else:
            self.receive_local_parameters()

    def update(self, use_gradients=True):
        """Update the global model

        Args:
            use_gradients (bool, optional): If True, update the global model with aggregated local gradients. Defaults to True.
        """
        if use_gradients:
            self.update_from_gradients()
        else:
            self.update_from_parameters()

    def _preprocess_local_gradients(self, uploaded_grad):
        return uploaded_grad

    def receive_local_gradients(self):
        """Receive local gradients"""
        self.uploaded_gradients = [
            self._preprocess_local_gradients(c.upload_gradients()) for c in self.clients
        ]

    def receive_local_parameters(self):
        """Receive local parameters"""
        self.uploaded_parameters = [c.upload_parameters() for c in self.clients]

    def update_from_gradients(self, weight=None):
        """Update the global model with the local gradients.

        Args:
            weight (list of float, optional): weight for each client. Defaults to None.
        """
        if weight is None:
            weight = np.ones(self.num_clients) / self.num_clients
            weight = weight.tolist()

        self.aggregated_gradients = [
            torch.zeros_like(params) for params in self.server_model.parameters()
        ]
        len_gradients = len(self.aggregated_gradients)

        for i, gradients in enumerate(self.uploaded_gradients):
            for gradient_id in range(len_gradients):
                self.aggregated_gradients[gradient_id] = (
                    gradients[gradient_id] * weight[i]
                    + self.aggregated_gradients[gradient_id]
                )

        if self.server_side_update:
            self.optimizer.step(self.aggregated_gradients)

    def update_from_parameters(self, weight=None):
        """Update the global model with the local model parameters.

        Args:
            weight (list of float, optional): weight for each client. Defaults to None.
        """
        if weight is None:
            weight = np.ones(self.num_clients) / self.num_clients

        averaged_params = self.uploaded_parameters[0]

        for k in averaged_params.keys():
            for i in range(0, len(self.uploaded_parameters)):
                local_model_params = self.uploaded_parameters[i]
                w = weight[i]
                if i == 0:
                    averaged_params[k] = local_model_params[k] * w
                else:
                    averaged_params[k] += local_model_params[k] * w

        self.server_model.load_state_dict(averaged_params)

    def distribute(self):
        """Distribute the current global model to each client.

        Args:
            force_send_model_state_dict (bool, optional): If True, send the global model as the dictionary of model state regardless of other parameters. Defaults to False.
        """
        for client in self.clients:
            if type(client) != int:
                if self.server_side_update or self.force_send_model_state_dict:
                    client.download(self.server_model.state_dict())
                else:
                    client.download(self.aggregated_gradients)


def attach_mpi_to_fedavgserver(cls):
    class MPIFedAVGServerWrapper(cls):
        """MPI Wrapper for FedAVG-based Server"""

        def __init__(self, comm, *args, **kwargs):
            self.comm = comm
            super(MPIFedAVGServerWrapper, self).__init__(*args, **kwargs)
            self.num_clients = len(self.clients)
            self.round = 0

        def action(self):
            self.receive()
            self.update()
            self.distribute()
            self.round += 1

        def receive(self):
            self.receive_local_gradients()

        def receive_local_gradients(self):
            self.uploaded_gradients = []

            while len(self.uploaded_gradients) < self.num_clients:
                gradients_received = self.comm.recv(tag=GRADIENTS_TAG)
                self.uploaded_gradients.append(
                    self._preprocess_local_gradients(gradients_received)
                )

        def distribute(self):
            for client_id in self.clients:
                self.comm.send(
                    self.server_model.state_dict(),
                    dest=client_id,
                    tag=PARAMETERS_TAG,
                )

        def mpi_initialize(self):
            self.distribute()

    return MPIFedAVGServerWrapper


class MPIFedAVGServerManager(BaseManager):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def attach(self, cls):
        return attach_mpi_to_fedavgserver(cls, *self.args, **self.kwargs)
